/******************************************************************************/
/*                                                                            */
/*  Cedocida DV-Codec                                                         */
/*                                                                            */
/*  Copyright (C) 2005 Andreas Dittrich <dittrich@eit.uni-kl.de>              */
/*  Project start: 15.12.2003                                                 */
/*  Last change  : 22.05.2005                                                 */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/*  This program is free software; you can redistribute it and/or modify      */
/*  it under the terms of the GNU General Public License as published by      */
/*  the Free Software Foundation; either version 2 of the License, or         */
/*  (at your option) any later version.                                       */
/*                                                                            */
/*  This program is distributed in the hope that it will be useful,           */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/*  GNU General Public License for more details.                              */
/*                                                                            */
/*  You should have received a copy of the GNU General Public License         */
/*  along with this program; if not, write to the Free Software               */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
/*                                                                            */
/******************************************************************************/


#include "codec_vfw.h"
#include <stdio.h>
#include <memory.h>

#include "inttypes.h"
#include "encode.h"
#include "decode.h"

#include "config.h"
#include "resource.h"


extern lpconfigvars_t lpmyconfig;


__declspec(dllexport) LRESULT WINAPI DriverProc(DWORD dwDriverID, HDRVR hDriver, UINT uiMessage, LPARAM lParam1, LPARAM lParam2) 
{

	cedocida_state_t *cedocida_state = (cedocida_state_t *)(UINT)dwDriverID; 
	ICOPEN *icinfo = (ICOPEN *)lParam2;
	switch (uiMessage) {

		/****************************************
			standard driver messages
		****************************************/

		case DRV_INSTALL:
		case DRV_REMOVE: return (LRESULT)DRV_OK;
		case DRV_ENABLE: return (LRESULT)1L;
		case DRV_DISABLE:
		case DRV_LOAD: return (LRESULT)1L;
		case DRV_OPEN:
			if (icinfo && icinfo->fccType != ICTYPE_VIDEO) return NULL;
			load_configuration( CONFIGURATION_FILENAME );
			cedocida_state = new cedocida_state_t;
			if (icinfo) 
				icinfo->dwError = ( cedocida_state != NULL ? ICERR_OK : ICERR_MEMORY );
			return (LRESULT)(DWORD)(UINT) cedocida_state;
		case DRV_CLOSE: return (LRESULT)1L;
		case DRV_FREE: return (LRESULT)1L;
		case DRV_QUERYCONFIGURE: return (LRESULT)1L;
		case DRV_CONFIGURE: return my_drv_configure(lParam1,lParam2);

		/****************************************
			state messages
		****************************************/

		case ICM_ABOUT: return my_icm_about(lParam1,lParam2);
		case ICM_CONFIGURE: return my_icm_configure(lParam1,lParam2);
		case ICM_GETSTATE: return my_icm_getstate(lParam1, lParam2);
		case ICM_SETSTATE: return my_icm_setstate(lParam1, lParam2);
		case ICM_GET: return 0;
		case ICM_GETQUALITY:
		case ICM_SETQUALITY:
		case ICM_GETDEFAULTQUALITY:
		case ICM_GETDEFAULTKEYFRAMERATE: return ICERR_UNSUPPORTED;
		case ICM_GETINFO: return my_icm_getinfo(lParam1, lParam2);

		/****************************************
			compress messages
		****************************************/

		case ICM_COMPRESS_QUERY: return my_icm_compress_query(lParam1, lParam2);
		case ICM_COMPRESS_GET_FORMAT: return my_icm_compress_get_format(lParam1, lParam2);
		case ICM_COMPRESS_GET_SIZE: return my_icm_get_size(lParam1, lParam2);
		case ICM_COMPRESS_BEGIN: return my_icm_compress_begin(lParam1, lParam2);
		case ICM_COMPRESS: return my_icm_compress(lParam1, lParam2);
		case ICM_COMPRESS_END: return my_icm_compress_end(lParam1, lParam2);

		/****************************************
			decompress messages
		****************************************/

		case ICM_DECOMPRESS_GET_PALETTE: return ICERR_BADFORMAT;
		case ICM_DECOMPRESS_SET_PALETTE: return ICERR_UNSUPPORTED;
		case ICM_DECOMPRESS_QUERY: return my_icm_decompress_query(lParam1, lParam2);
		case ICM_DECOMPRESS_GET_FORMAT: return my_icm_decompress_get_format(lParam1, lParam2);
		case ICM_DECOMPRESS_BEGIN: return my_icm_decompress_begin(lParam1, lParam2);
		case ICM_DECOMPRESS: return my_icm_decompress(lParam1, lParam2);
		case ICM_DECOMPRESS_END: return my_icm_decompress_end(lParam1, lParam2);

	}

	if (uiMessage < DRV_USER)
		return DefDriverProc(dwDriverID, hDriver, uiMessage, lParam1, lParam2);
	else
		return ICERR_UNSUPPORTED;
}


HMODULE hmodule_cedocida=0;

BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD, LPVOID) {
  hmodule_cedocida = (HMODULE)hinstDLL;
  return TRUE;
}



/****************************************
	helper functions
****************************************/

static bool is_YV12(LPBITMAPINFOHEADER lpbiIn)
{
	if (!lpbiIn) return 0;
	return ( (lpbiIn->biCompression == FOURCC_YV12) );
}

static bool is_YUY2(LPBITMAPINFOHEADER lpbiIn)
{
	if (!lpbiIn) return 0;
	return ( (lpbiIn->biCompression == FOURCC_YUY2) || (lpbiIn->biCompression == FOURCC_VYUY) );
}

static bool is_DVSD(LPBITMAPINFOHEADER lpbiIn)
{
	if (!lpbiIn) return 0;
	return ( lpbiIn->biCompression == FOURCC_DVSD );
}

static bool is_RGB24(LPBITMAPINFOHEADER lpbiIn)
{
	if (!lpbiIn) return 0;

	const int fourcc = lpbiIn->biCompression;
	const int bitcount = lpbiIn->biBitCount;

	return (  ( fourcc == 0 || fourcc == ' BID' ) && ( bitcount == 24)  );
}

static bool is_RGB32(LPBITMAPINFOHEADER lpbiIn)
{
	if (!lpbiIn) return 0;

	const int fourcc = lpbiIn->biCompression;
	const int bitcount = lpbiIn->biBitCount;

	return (  ( fourcc == 0 || fourcc == ' BID' ) && ( bitcount == 32)  );
}


int get_colorformat(LPBITMAPINFOHEADER lpbiIn)
{
	int fourcc;
	int bitcount;

	if (!lpbiIn) return 0;

	/* rgb only: negative height specifies top down image */
	if ( lpbiIn->biHeight < 0 ) return COLORFORMAT_UNKNOWN;

	fourcc = lpbiIn->biCompression;
	bitcount = lpbiIn->biBitCount;

	if (  ( fourcc == 0 || fourcc == ' BID' ) && ( bitcount == 32)  ) 
		return COLORFORMAT_RGB32;
	else if (  ( fourcc == 0 || fourcc == ' BID' ) && ( bitcount == 24)  ) 
		return COLORFORMAT_RGB24;
	else if ( fourcc == FOURCC_YUY2 ) 
		return COLORFORMAT_YUY2;
	else if ( fourcc == FOURCC_YV12 ) 
		return COLORFORMAT_YV12;
	else
		return COLORFORMAT_UNKNOWN;
}


int get_input_colorformat(LPBITMAPINFOHEADER lpbiIn)
{
	int color_format = get_colorformat(lpbiIn);

	if ( lpmyconfig->encoder_force_input_format )
	{
		if (color_format==lpmyconfig->encoder_input_format) return color_format;
		else if ( (lpmyconfig->encoder_input_format==COLORFORMAT_RGB32) & 
			(color_format==COLORFORMAT_RGB24) ) return color_format;
		else return COLORFORMAT_NA;
	}

	return color_format;
}


int get_output_colorformat(LPBITMAPINFOHEADER lpbiIn)
{
	int color_format = get_colorformat(lpbiIn);

	if ( lpmyconfig->decoder_force_output_format )
	{
		if (color_format==lpmyconfig->decoder_output_format) return color_format;
		else if ( (lpmyconfig->decoder_output_format==COLORFORMAT_RGB32) & 
			(color_format==COLORFORMAT_RGB24) ) return color_format;
		else return COLORFORMAT_NA;
	}

	return color_format;
}


/****************************************
	standard driver messages
****************************************/

BOOL CALLBACK ConfigureDialogProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);

DWORD my_drv_configure(LPARAM lParam1, LPARAM lParam2)
{
	HWND hwnd = (HWND)lParam1;
	DialogBox(hmodule_cedocida, MAKEINTRESOURCE(IDD_DIALOG_CONFIGURE), hwnd, ConfigureDialogProc);
	return DRV_OK;
}


/****************************************
	state messages
****************************************/

DWORD my_icm_about(LPARAM lParam1, LPARAM lParam2)
{
	return ICERR_UNSUPPORTED;
}

DWORD my_icm_configure(LPARAM lParam1, LPARAM lParam2)
{
	if (lParam1 == -1)
		return ICERR_OK;
	else 
		my_drv_configure(lParam1, lParam2);

	return ICERR_OK;
}

DWORD my_icm_getstate(LPARAM lParam1, LPARAM lParam2)
{
	return 0;
}

DWORD my_icm_setstate(LPARAM lParam1, LPARAM lParam2)
{
	return 0;
}

DWORD my_icm_getinfo(LPARAM lParam1, LPARAM lParam2)
{
	ICINFO* icinfo = (ICINFO *)lParam1;
	DWORD dwSize = lParam2;

	if (icinfo == NULL) return sizeof(ICINFO);
	if (dwSize < sizeof(ICINFO)) return 0;

	icinfo->dwSize            = sizeof(ICINFO);
	icinfo->fccType           = ICTYPE_VIDEO;
	icinfo->fccHandler        = FOURCC_DVSD;
	icinfo->dwFlags           = 0;

	icinfo->dwVersion         = 0;
	icinfo->dwVersionICM      = ICVERSION;
	wcscpy(icinfo->szName, CODEC_NAME_L); 
	wcscpy(icinfo->szDescription, CODEC_DESCRIPTION_L);

	return lParam2;
}

/****************************************
	compress messages
****************************************/

DWORD my_icm_compress_query(LPARAM lParam1, LPARAM lParam2)
{
	LPBITMAPINFOHEADER lpbiIn = (LPBITMAPINFOHEADER)lParam1;
	LPBITMAPINFOHEADER lpbiOut = (LPBITMAPINFOHEADER)lParam2;
	
	if ( get_input_colorformat(lpbiIn) > 0 )
	{
		if (!lpbiOut) 
			return ICERR_OK;
		else 
		{
			if ( is_DVSD(lpbiOut) )
				return ICERR_OK;
			else 
				return ICERR_BADFORMAT;
		}
	}
	else
	{
		return ICERR_BADFORMAT;
	}
}

DWORD my_icm_compress_get_format(LPARAM lParam1, LPARAM lParam2)
{
	BOOL add_DVINFO = FALSE;

	LPBITMAPINFOHEADER lpbiIn = (LPBITMAPINFOHEADER)lParam1;
	LPBITMAPINFOHEADER lpbiOut = (LPBITMAPINFOHEADER)lParam2;

	if ( my_icm_compress_query(lParam1, NULL) != ICERR_OK )
		return ICERR_BADFORMAT;
  	
	if (!lpbiOut) return sizeof(BITMAPINFOHEADER) + ( add_DVINFO ? sizeof(DVINFO) : 0 );

	lpbiOut->biSize = sizeof(BITMAPINFOHEADER) + ( add_DVINFO ? sizeof(DVINFO) : 0 );
	lpbiOut->biWidth = 720;
	lpbiOut->biHeight = ( lpbiIn->biHeight==480 ? 480 : 576 );
	lpbiOut->biPlanes = 1;
	lpbiOut->biBitCount = 24;
	lpbiOut->biCompression = FOURCC_DVSD;
    lpbiOut->biSizeImage = 150*80*( lpbiIn->biHeight==480 ? 10 : 12 ); // 1 DV-Frame = 12 (or 10 for NTSC) DIF-Sequence  150 DIF-Blocks  80 Bytes = 144000 Bytes (or 12000 for NTSC)
    lpbiOut->biXPelsPerMeter = 0; //117*100; 
    lpbiOut->biYPelsPerMeter = 0; //128*100; 
    lpbiOut->biClrUsed = 0; 
    lpbiOut->biClrImportant = 0; 

	if (add_DVINFO)
	{
		DVINFO *lpdvinfoOut = (DVINFO*)((unsigned char*)lpbiOut + sizeof(BITMAPINFOHEADER));
		lpdvinfoOut->dwDVAAuxSrc = 0;
		lpdvinfoOut->dwDVAAuxCtl = 0;
		lpdvinfoOut->dwDVAAuxSrc1 = 0;
		lpdvinfoOut->dwDVAAuxCtl1 = 0;
		lpdvinfoOut->dwDVVAuxSrc = 0;
		lpdvinfoOut->dwDVVAuxCtl = 0;
		lpdvinfoOut->dwDVReserved[1] = 0;
		lpdvinfoOut->dwDVReserved[2] = 0;
	}

	return ICERR_OK;
}

DWORD my_icm_get_size(LPARAM lParam1, LPARAM lParam2)
{
	LPBITMAPINFOHEADER lpbiIn = (LPBITMAPINFOHEADER)lParam1;
	LPBITMAPINFOHEADER lpbiOut = (LPBITMAPINFOHEADER)lParam2;

	return 150*80*( lpbiIn->biHeight==480 ? 10 : 12 ); // 1 DV-Frame = 12 (or 10 for NTSC) DIF-Sequence  150 DIF-Blocks  80 Bytes = 144000 Bytes (or 12000 for NTSC)
}

DWORD my_icm_compress_begin(LPARAM lParam1, LPARAM lParam2)
{
	LPBITMAPINFOHEADER lpbiIn = (LPBITMAPINFOHEADER)lParam1;
	LPBITMAPINFOHEADER lpbiOut = (LPBITMAPINFOHEADER)lParam2;

	if ( (get_input_colorformat(lpbiIn)>0) && 
		 (lpbiIn->biWidth==720) && 
		 ((lpbiIn->biHeight==480) || (lpbiIn->biHeight==576) )&& 
		 (is_DVSD(lpbiOut)) )
		return ICERR_OK;
	else
		return ICERR_BADFORMAT;
}

DWORD my_icm_compress(LPARAM lParam1, LPARAM lParam2)
{
	ICCOMPRESS *icinfo = (ICCOMPRESS *)lParam1;
	DWORD dwSize = (DWORD)lParam2;
	
	if (icinfo->lpckid) *icinfo->lpckid = FOURCC_DVSD;
	*icinfo->lpdwFlags = AVIIF_KEYFRAME;

	if ( my_icm_compress_begin( (LPARAM)icinfo->lpbiInput, (LPARAM)icinfo->lpbiOutput ) == ICERR_OK )
	{
		unsigned int *in = (unsigned int *)icinfo->lpInput;
		unsigned int *out = (unsigned int *)icinfo->lpOutput;

		char color_format = (char)get_input_colorformat(icinfo->lpbiInput);
		char quality_mode = 0;

		if ( icinfo->lpbiInput->biHeight==480 )
		{
			encode_NTSC_DVSD( in, out, color_format );
			icinfo->lpbiOutput->biSizeImage = 120000;
		}
		else
		{
			encode_PAL_DVSD( in, out, color_format );
			icinfo->lpbiOutput->biSizeImage = 144000;
		}

		return ICERR_OK;
	}
	else return ICERR_BADFORMAT;
}

DWORD my_icm_compress_end(LPARAM lParam1, LPARAM lParam2)
{
	return ICERR_OK;
}

/****************************************
	decompress messages
****************************************/

DWORD my_icm_decompress_query(LPARAM lParam1, LPARAM lParam2)
{
	LPBITMAPINFOHEADER lpbiIn = (LPBITMAPINFOHEADER)lParam1;
	LPBITMAPINFOHEADER lpbiOut = (LPBITMAPINFOHEADER)lParam2;
	
	if ( is_DVSD(lpbiIn) )
	{
		if (!lpbiOut) 
			return ICERR_OK;
		else if ( get_output_colorformat(lpbiOut) > 0 ) 
			return ICERR_OK;
		else 
			return ICERR_BADFORMAT;
	}
	else return ICERR_BADFORMAT;
}

#define ABS( x ) ( ((x)<0) ? -(x) : (x) )

DWORD my_icm_decompress_get_format(LPARAM lParam1, LPARAM lParam2)
{
	LPBITMAPINFOHEADER lpbiIn = (LPBITMAPINFOHEADER)lParam1;
	LPBITMAPINFOHEADER lpbiOut = (LPBITMAPINFOHEADER)lParam2;

	if (lpbiOut == NULL) return sizeof(BITMAPINFOHEADER);

	if (!is_DVSD(lpbiIn)) return ICERR_BADFORMAT;

	memcpy( lpbiOut, lpbiIn, sizeof(BITMAPINFOHEADER) );

	if ( lpmyconfig->decoder_output_format == COLORFORMAT_YV12 )
	{	// YV12
		lpbiOut->biSize = sizeof(BITMAPINFOHEADER);
		lpbiOut->biPlanes = 1;
		lpbiOut->biBitCount = 12;       
		lpbiOut->biCompression = FOURCC_YV12;
		lpbiOut->biSizeImage = (lpbiIn->biWidth * ABS(lpbiIn->biHeight) * 3)/2;
	}
	else if ( lpmyconfig->decoder_output_format == COLORFORMAT_RGB32 )
	{   // RGB32
		lpbiOut->biSize = sizeof(BITMAPINFOHEADER);
		lpbiOut->biPlanes = 1;
		lpbiOut->biBitCount = 32;    
		lpbiOut->biCompression = ' BID';
		lpbiOut->biSizeImage = lpbiIn->biWidth * ABS(lpbiIn->biHeight) * 4;
	}
	else
	{	// YUY2
		lpbiOut->biSize = sizeof(BITMAPINFOHEADER);
		lpbiOut->biPlanes = 1;
		lpbiOut->biBitCount = 16;       // YUY2
		lpbiOut->biCompression = FOURCC_YUY2;
		lpbiOut->biSizeImage = lpbiIn->biWidth * ABS(lpbiIn->biHeight) * 2;
	}

	
	return ICERR_OK;
}

DWORD my_icm_decompress_begin(LPARAM lParam1, LPARAM lParam2)
{
	LPBITMAPINFOHEADER lpbiIn = (LPBITMAPINFOHEADER)lParam1;
	LPBITMAPINFOHEADER lpbiOut = (LPBITMAPINFOHEADER)lParam2;

	if ( (is_DVSD(lpbiIn)) && (get_output_colorformat(lpbiOut)>0) && 
		 (lpbiOut->biWidth==720) && ((lpbiOut->biHeight==480)||(lpbiOut->biHeight==576)) )
		return ICERR_OK;
	else
		return ICERR_BADFORMAT;
}

DWORD my_icm_decompress(LPARAM lParam1, LPARAM lParam2)
{
	ICDECOMPRESS *icinfo = (ICDECOMPRESS *)lParam1; 
	DWORD dwSize = (DWORD)lParam2;
	
	if ( my_icm_decompress_begin((LPARAM)icinfo->lpbiInput, (LPARAM)icinfo->lpbiOutput) == ICERR_OK )
	{
		unsigned char *in = (unsigned char *)icinfo->lpInput;
		unsigned int *out = (unsigned int *)icinfo->lpOutput;

		char color_format = (char)get_output_colorformat(icinfo->lpbiOutput);

		if ( icinfo->lpbiOutput->biHeight==480 )
			decode_NTSC_DVSD( in, out, color_format );
		else
			decode_PAL_DVSD( in, out, color_format );

		return ICERR_OK;
  }
  else return ICERR_BADFORMAT;
}

DWORD my_icm_decompress_end(LPARAM lParam1, LPARAM lParam2)
{
	return ICERR_OK;
}


