/******************************************************************************/
/*                                                                            */
/*  Cedocida DV-Codec                                                         */
/*                                                                            */
/*  Copyright (C) 2005 Andreas Dittrich <dittrich@eit.uni-kl.de>              */
/*  Project start: 15.12.2003                                                 */
/*  Last change  : 18.06.2006                                                 */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/*  This program is free software; you can redistribute it and/or modify      */
/*  it under the terms of the GNU General Public License as published by      */
/*  the Free Software Foundation; either version 2 of the License, or         */
/*  (at your option) any later version.                                       */
/*                                                                            */
/*  This program is distributed in the hope that it will be useful,           */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/*  GNU General Public License for more details.                              */
/*                                                                            */
/*  You should have received a copy of the GNU General Public License         */
/*  along with this program; if not, write to the Free Software               */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
/*                                                                            */
/******************************************************************************/


#include <memory.h>

#include "inttypes.h"

#include "signalproc.h"
#include "dv_spec.h"
#include "dct/dct.h"
#include "vlc/decode_vlc.h"
#include "decode_tables.h"
#include "decode.h"
#include "config.h"

#include "convert/convert.h"


#define LIMIT_1_TO_254( x ) ( ((x)<1) ? 1 : ( ((x)>254) ? 254 : (x) ) )
#define LIMIT_2_TO_508( x ) ( ((x)<2) ? 2 : ( ((x)>508) ? 508 : (x) ) )
#define LIMIT_MINUS255_TO_PLUS255( x ) ( ((x)<-255) ? (-255) : ( ((x)>255) ? (255) : (x) ) )


// fixed size temporary storage space
extern int16_t buffer_5_MBs[5*6*64];	
extern uint32_t buffer_frame[720*576];	// for quick and dirty color-conversion

extern lpconfigvars_t lpmyconfig;




void decode_PAL_DVSD( uint8_t *src_DV, uint32_t *dst, char color_format )
{
	int i, r, j, n;
	int16_t *dst_5_MBs = buffer_5_MBs;
	uint32_t *dst_step1 = dst;
	uint32_t *dst_step2 = buffer_frame;

	if (color_format==COLORFORMAT_YUY2)
	{
		dst_step1 = buffer_frame;
	}

	for (i=0; i<12; i++)	// sequence
	{
		for (r=0; r<3; r++)	// row_MB in SB
		{
			for (j=0; j<9; j++)	// col_MB in SB
			{
				int block_indx = 5*SB_row_order[r][j]; 
				int DBN_1st = (7+block_indx+block_indx/15);	// first DIF_blocknumber
				uint8_t *src_DIF_videosegment = src_DV + i*(150*80) + DBN_1st*(80);

				decode_DIF_videosegment( src_DIF_videosegment, dst_5_MBs, (uint8_t)i, DBN_1st );

				for (n=0; n<5; n++)		// col_SB
				{
					// save it first as YV12 and then convert to other formats
					// todo: direct conversion to the right format
					uint32_t *src = (uint32_t*)( dst_5_MBs + n*6*64 );

					uint32_t *dst_Y = dst_step1; int dst_Y_pitch = (720/4);
					uint32_t *dst_V = dst_Y + 576*dst_Y_pitch; int dst_V_pitch = (360/4);
					uint32_t *dst_U = dst_V + 288*dst_V_pitch; int dst_U_pitch = dst_V_pitch;

					const static int modulo_offset[5] = { 2, 6, 8, 0, 4};
					const static int j_SB[5] = { 2, 1, 3, 0, 4};

					int row_SB = (i+modulo_offset[n])%12;

					uint32_t *dst_YUV420_Y= dst_Y  + (j_SB[n])*(9*16/4) + j*(16/4) + (row_SB*3+r)*16*dst_Y_pitch;
					uint32_t *dst_YUV420_V = dst_V + (j_SB[n])*(9*8/4)  + j*(8/4) + (row_SB*3+r)*8*dst_V_pitch;
					uint32_t *dst_YUV420_U = dst_U + (j_SB[n])*(9*8/4)  + j*(8/4) + (row_SB*3+r)*8*dst_U_pitch;


					deinterleave_MB_YV12_PAL(src, dst_YUV420_Y, dst_Y_pitch, 
						dst_YUV420_U, dst_U_pitch, dst_YUV420_V, dst_V_pitch );

				}
			}
		}
	}


	if (color_format==COLORFORMAT_YV12)
	{
		if (lpmyconfig->decoder_YV12_chroma_sampling!=YV12_CHROMASAMPLING_DV)
		{
			YUV420_to_YUY2( (uint8_t*)dst_step1, 720, 
				(uint8_t*)(dst_step1+576*720/4+288*90), 360, 
				(uint8_t*)(dst_step1+576*720/4), 360,
				(uint8_t*)dst_step2, 720*2, 
				720, 576 );

			if (lpmyconfig->decoder_YV12_chroma_sampling==YV12_CHROMASAMPLING_MPEG2_I)
			{
				mmx_yuy2_i_to_yv12( (BYTE*)dst_step2, 720*2, 720*2, 
					(BYTE*)dst, (BYTE*)dst+576*720+288*360, (BYTE*)dst+576*720, 720, 360,
					576 );
			}
			else
			{
				mmx_yuy2_to_yv12( (BYTE*)dst_step2, 720*2, 720*2, 
					(BYTE*)dst, (BYTE*)dst+576*720+288*360, (BYTE*)dst+576*720, 720, 360,
					576 );
			}
		}
	}
	else if (color_format==COLORFORMAT_YUY2)
	{
		YUV420_to_YUY2( (uint8_t*)dst_step1, 720, 
			(uint8_t*)(dst_step1+576*720/4+288*90), 360, 
			(uint8_t*)(dst_step1+576*720/4), 360,
			(uint8_t*)dst, 720*2, 
			720, 576 );
	}
	else if (color_format==COLORFORMAT_RGB32)
	{
		YUV420_to_YUY2( (uint8_t*)dst_step1, 720, 
			(uint8_t*)(dst_step1+576*720/4+288*90), 360, 
			(uint8_t*)(dst_step1+576*720/4), 360,
			(uint8_t*)dst_step2, 720*2, 
			720, 576 );

		if (lpmyconfig->decoder_YUV_to_RGB_conversion==RGB2YUV_601)
			mmx_YUY2toRGB32( (BYTE*)dst_step2, (BYTE*)dst, (BYTE*)dst_step2+720*576*2, 720*2, 720*2, 0);
		else
			mmx_YUY2toRGB32( (BYTE*)dst_step2, (BYTE*)dst, (BYTE*)dst_step2+720*576*2, 720*2, 720*2, 1);
	}
	else if (color_format==COLORFORMAT_RGB24)
	{
		YUV420_to_YUY2( (uint8_t*)dst_step1, 720, 
			(uint8_t*)(dst_step1+576*720/4+288*90), 360, 
			(uint8_t*)(dst_step1+576*720/4), 360,
			(uint8_t*)dst_step2, 720*2, 
			720, 576 );

		if (lpmyconfig->decoder_YUV_to_RGB_conversion==RGB2YUV_601)
			mmx_YUY2toRGB24( (BYTE*)dst_step2, (BYTE*)dst, (BYTE*)dst_step2+720*576*2, 720*2, 720*2, 0);
		else
			mmx_YUY2toRGB24( (BYTE*)dst_step2, (BYTE*)dst, (BYTE*)dst_step2+720*576*2, 720*2, 720*2, 1);

	}
}


void decode_NTSC_DVSD( uint8_t *src_DV, uint32_t *dst, char color_format )
{
	const int col_SB[5] = { 2, 1, 3, 0, 4};
	const int row_SB[10*5] = { 
		2, 3, 4, 5, 6, 7, 8, 9, 0, 1,
		6, 7, 8, 9, 0, 1, 2, 3, 4, 5,
		8, 9, 0, 1, 2, 3, 4, 5, 6, 7,
		0, 1, 2, 3, 4, 5, 6, 7, 8, 9,
		4, 5, 6, 7, 8, 9, 0, 1, 2, 3
	};

	int i, j, k, p;
	int16_t *dst_5_MBs = buffer_5_MBs;
	uint32_t *dst_step1 = dst;
	uint32_t *dst_step2 = buffer_frame;

	if (color_format==COLORFORMAT_YUY2)
	{
		dst_step1 = buffer_frame;
	}

	for (i=0; i<10; i++)	// sequence
	{
		for (k=0; k<27; k++)	// MB-index in SB
		{			
			int block_indx = 5*k; 
			int DBN_1st = (7+block_indx+block_indx/15);	// first DIF_blocknumber
			uint8_t *src_DIF_videosegment = src_DV + i*(150*80) + DBN_1st*(80);

			decode_DIF_videosegment( src_DIF_videosegment, dst_5_MBs, (uint8_t)i, DBN_1st );
			
			for(j=0; j<5; j++)	// superblock row number: a,b,c,d,e
			{
				p = col_SB[j];

				// save it first as YV12 and then convert to other formats
				// todo: direct conversion to the right format
				uint32_t *src = (uint32_t*)( dst_5_MBs + j*6*64 );

				int col_MB = MB_pos_NTSC[p][2*k];
				int row_MB = MB_pos_NTSC[p][2*k+1] + row_SB[j*10+i]*6;

				uint32_t *dst_y = dst_step1 + col_MB*(4*8/4) + row_MB*(8*720/4);
				uint32_t *dst_v = dst_step1 + (720*480/4) + col_MB*(8/4) + row_MB*(8*180/4);
				uint32_t *dst_u = dst_step1 + (720*480/4 + 180*480/4) + col_MB*(8/4) + row_MB*(8*180/4);
				int dst_y_pitch = 720;
				int dst_v_pitch = 180;
				int dst_u_pitch = dst_v_pitch;

				if ( (p==4) && (k>23) )
				{
					deinterleave_MBborder_YV12_NTSC(src, dst_y, dst_y_pitch, 
						dst_u, dst_u_pitch, dst_v, dst_v_pitch );
				}
				else
				{
					deinterleave_MB_YV12_NTSC(src, dst_y, dst_y_pitch, 
						dst_u, dst_u_pitch, dst_v, dst_v_pitch );
				}			
			}
		}
	}


	if (color_format==COLORFORMAT_YV12)
	{
		if (lpmyconfig->decoder_YV12_chroma_sampling!=YV12_CHROMASAMPLING_DV)
		{
			YUV411_to_YUY2( (uint8_t*)dst_step1, 720, 
				(uint8_t*)(dst_step1+480*720/4+480*180/4), 180, 
				(uint8_t*)(dst_step1+480*720/4), 180,
				(uint8_t*)dst_step2, 720*2, 
				720, 480 );

			if (lpmyconfig->decoder_YV12_chroma_sampling==YV12_CHROMASAMPLING_MPEG2_I)
			{
				mmx_yuy2_i_to_yv12( (BYTE*)dst_step2, 720*2, 720*2, 
					(BYTE*)dst, (BYTE*)dst+480*720+240*360, (BYTE*)dst+480*720, 720, 360,
					480 );
			}
			else
			{
				mmx_yuy2_to_yv12( (BYTE*)dst_step2, 720*2, 720*2, 
					(BYTE*)dst, (BYTE*)dst+480*720+240*360, (BYTE*)dst+480*720, 720, 360,
					480 );
			}
		}
	}
	else if (color_format==COLORFORMAT_YUY2)
	{
		YUV411_to_YUY2( (uint8_t*)dst_step1, 720, 
			(uint8_t*)(dst_step1+480*720/4+480*180/4), 180, 
			(uint8_t*)(dst_step1+480*720/4), 180,
			(uint8_t*)dst, 720*2, 
			720, 480 );
	}
	else if (color_format==COLORFORMAT_RGB32)
	{
		YUV411_to_YUY2( (uint8_t*)dst_step1, 720, 
			(uint8_t*)(dst_step1+480*720/4+480*180/4), 180, 
			(uint8_t*)(dst_step1+480*720/4), 180,
			(uint8_t*)dst_step2, 720*2, 
			720, 480 );

		if (lpmyconfig->decoder_YUV_to_RGB_conversion==RGB2YUV_601)
			mmx_YUY2toRGB32( (BYTE*)dst_step2, (BYTE*)dst, (BYTE*)dst_step2+720*480*2, 720*2, 720*2, 0);
		else
			mmx_YUY2toRGB32( (BYTE*)dst_step2, (BYTE*)dst, (BYTE*)dst_step2+720*480*2, 720*2, 720*2, 1);
	}
	else if (color_format==COLORFORMAT_RGB24)
	{
		YUV411_to_YUY2( (uint8_t*)dst_step1, 720, 
			(uint8_t*)(dst_step1+480*720/4+480*180/4), 180, 
			(uint8_t*)(dst_step1+480*720/4), 180,
			(uint8_t*)dst_step2, 720*2, 
			720, 480 );

		if (lpmyconfig->decoder_YUV_to_RGB_conversion==RGB2YUV_601)
			mmx_YUY2toRGB24( (BYTE*)dst_step2, (BYTE*)dst, (BYTE*)dst_step2+720*480*2, 720*2, 720*2, 0);
		else
			mmx_YUY2toRGB24( (BYTE*)dst_step2, (BYTE*)dst, (BYTE*)dst_step2+720*480*2, 720*2, 720*2, 1);

	}
}


void decode_DIF_videosegment( uint8_t *DIF_block, int16_t *dst_5_MBs, 
								 uint32_t DIF_sequence_number, uint32_t DIF_block_number )
{
	static uint32_t DCT_mode[5*6];
	static uint32_t quantization_number[5]; 
	static uint32_t class_number[5*6]; 
	int j, l, n;
	int16_t *block = dst_5_MBs;

	int iresult;

	iresult = decode_videosegment( dst_5_MBs, quantization_number, DCT_mode, class_number, 
									   DIF_block, DIF_sequence_number, DIF_block_number );
	n = 0;
	for (j=0; j<5; j++)
	{
		uint32_t QNO = quantization_number[j];

		for (l=0; l<6; l++)
		{
			int c1c0 = class_number[n];
			int mo = DCT_mode[n];
			int q_code = quantization_code_table_extended[(c1c0<<4)+QNO];
			
			if (mo==0)
			{
				int16_t *q_step = (int16_t*)quantization_step_88_extended + q_code*64;

				// todo: use and test mmx-version
				multiply_clip_16bit( block, block, q_step );
				block[0] = block[0]*16;
				weighting_block_MMX( block, block, (int16_t*)W_88_inv_16bit );

				idct_88( block, block);

				int k;
				for(k=0;k<64;k++) 
				{
					block[k] += 128;
					if (block[k]<0) block[k]=0;
					else if (block[k]>255) block[k]=255;
				}
			}
			else
			{
				int16_t *q_step = (int16_t*)quantization_step_248_extended + q_code*64;
				// todo: use and test mmx-version
				multiply_clip_16bit( block, block, q_step );
				block[0] = block[0]*16;
				weighting_block_MMX( block, block, (int16_t*)W_248_inv_16bit );

				idct_248( block, block);

				int k;
				for(k=0;k<64;k++) 
				{
					block[k] += 128;
					if (block[k]<0) block[k]=0;
					else if (block[k]>255) block[k]=255;
				}
			}

			n++;
			block += 64;
		}
	}
}


void deinterleave_MB_YV12_PAL(uint32_t *src, uint32_t *dst_YUV420_Y, int dst_YUV420_Y_pitch, 
						  uint32_t *dst_YUV420_U, int dst_YUV420_U_pitch, 
					      uint32_t *dst_YUV420_V, int dst_YUV420_V_pitch
					      )
{
	uint8_t *x = (uint8_t*)dst_YUV420_Y;
	int16_t *y = (int16_t*)src;
	int x_pitch = dst_YUV420_Y_pitch*4;
	int i,j;

	for( i=0; i<8; i++)
	{
		for( j=0; j<8; j++)
		{
			// Block_Y0
			x[j] = (uint8_t)y[i*8+j];
			// Block_Y1
			x[j+8] = (uint8_t)y[i*8+64+j];
		}
		x=x+x_pitch;
	}

	for( i=0; i<8; i++)
	{
		for( j=0; j<8; j++)
		{
			// Block_Y0
			x[j] = (uint8_t)y[i*8+128+j];
			// Block_Y1
			x[j+8] = (uint8_t)y[i*8+192+j];
		}
		x=x+x_pitch;
	}

	x = (uint8_t*)dst_YUV420_V;
	x_pitch = dst_YUV420_U_pitch*4;
	for( i=0; i<8; i++)
	{
		for( j=0; j<8; j++)
		{
			// Block_V
			x[j] = (uint8_t)y[i*8+256+j];
		}
		x=x+x_pitch;
	}

	x = (uint8_t*)dst_YUV420_U;
	x_pitch = dst_YUV420_V_pitch*4;
	for( i=0; i<8; i++)
	{
		for( j=0; j<8; j++)
		{
			// Block_V
			x[j] = (uint8_t)y[i*8+320+j];
		}
		x=x+x_pitch;
	}
}


void YUV420_to_YUY2( uint8_t* src_YUV420_Y, int src_YUV420_Y_pitch, 
					 uint8_t* src_YUV420_U, int src_YUV420_U_pitch, 
					 uint8_t* src_YUV420_V, int src_YUV420_V_pitch,
					 uint8_t* dst_YUY2, int dst_YUY2_pitch, int width, int height )
{
	int width_half = width/2;
	int height_minus_2 = height - 2;
	int x, y;
	int V_pitch_x2 = src_YUV420_V_pitch*2;
	int U_pitch_x2 = src_YUV420_U_pitch*2;

	uint8_t* out = dst_YUY2;
	uint8_t* Y = src_YUV420_Y;
	uint8_t* U = src_YUV420_U;
	uint8_t* V = src_YUV420_V;


	for( x=0; x<width_half; x++ )
	{
		out[0] = Y[0];
		out[1] = U[0];
		out[2] = Y[1];
		out[3] = V[0];
		out+=4;
		Y+=2;
		U++;
		V++;
	}
	out = out - 4*width_half + dst_YUY2_pitch;
	Y = Y - 2*width_half + src_YUV420_Y_pitch;
	U = U - width_half + src_YUV420_U_pitch;
	V = V - width_half + src_YUV420_V_pitch;

	for( x=0; x<width_half; x++ )
	{
		out[0] = Y[0];
		out[1] = U[0];
		out[2] = Y[1];
		out[3] = V[0];
		out+=4;
		Y+=2;
		U++;
		V++;
	}
	out = out - 4*width_half + dst_YUY2_pitch;
	Y = Y - 2*width_half + src_YUV420_Y_pitch;
	U = U - width_half - src_YUV420_U_pitch;
	V = V - width_half - src_YUV420_V_pitch;


	for ( y=2; y<height_minus_2; y+=4 )
	{

		for( x=0; x<width_half; x++ )
		{
			out[0] = Y[0];
			out[1] = U[0];
			out[2] = Y[1];
			out[3] = (uint8_t)( ((int)V[0]+(int)V[V_pitch_x2]+1)>>1 );
			out+=4;
			Y+=2;
			U++;
			V++;
		}
		out = out - 4*width_half + dst_YUY2_pitch;
		Y = Y - 2*width_half + src_YUV420_Y_pitch;
		U = U - width_half + src_YUV420_U_pitch;
		V = V - width_half + src_YUV420_V_pitch;

		for( x=0; x<width_half; x++ )
		{
			out[0] = Y[0];
			out[1] = U[0];
			out[2] = Y[1];
			out[3] = (uint8_t)( ((int)V[0]+(int)V[V_pitch_x2]+1)>>1 );
			out+=4;
			Y+=2;
			U++;
			V++;
		}
		out = out - 4*width_half + dst_YUY2_pitch;
		Y = Y - 2*width_half + src_YUV420_Y_pitch;
		U = U - width_half + src_YUV420_U_pitch;
		V = V - width_half + src_YUV420_V_pitch;

		for( x=0; x<width_half; x++ )
		{
			out[0] = Y[0];
			out[1] = (uint8_t)( ((int)U[-U_pitch_x2]+(int)U[0]+1)>>1 );
			out[2] = Y[1];
			out[3] = V[0];
			out+=4;
			Y+=2;
			U++;
			V++;
		}
		out = out - 4*width_half + dst_YUY2_pitch;
		Y = Y - 2*width_half + src_YUV420_Y_pitch;
		U = U - width_half + src_YUV420_U_pitch;
		V = V - width_half + src_YUV420_V_pitch;

		for( x=0; x<width_half; x++ )
		{
			out[0] = Y[0];
			out[1] = (uint8_t)( ((int)U[-U_pitch_x2]+(int)U[0]+1)>>1 );
			out[2] = Y[1];
			out[3] = V[0];
			out+=4;
			Y+=2;
			U++;
			V++;
		}
		out = out - 4*width_half + dst_YUY2_pitch;
		Y = Y - 2*width_half + src_YUV420_Y_pitch;
		U = U - width_half - src_YUV420_U_pitch;
		V = V - width_half - src_YUV420_V_pitch;
	}


	for( x=0; x<width_half; x++ )
	{
		out[0] = Y[0];
		out[1] = U[0];
		out[2] = Y[1];
		out[3] = V[0];
		out+=4;
		Y+=2;
		U++;
		V++;
	}
	out = out - 4*width_half + dst_YUY2_pitch;
	Y = Y - 2*width_half + src_YUV420_Y_pitch;
	U = U - width_half + src_YUV420_U_pitch;
	V = V - width_half + src_YUV420_V_pitch;

	for( x=0; x<width_half; x++ )
	{
		out[0] = Y[0];
		out[1] = U[0];
		out[2] = Y[1];
		out[3] = V[0];
		out+=4;
		Y+=2;
		U++;
		V++;
	}

}




void deinterleave_MB_YV12_NTSC(uint32_t *src, uint32_t *dst_YUV420_Y, int dst_YUV420_Y_pitch, 
						  uint32_t *dst_YUV420_U, int dst_YUV420_U_pitch, 
					      uint32_t *dst_YUV420_V, int dst_YUV420_V_pitch
					      )
{
	int n, row, col;
	uint8_t *dst_y = (uint8_t *)dst_YUV420_Y;
	uint8_t *dst_u = (uint8_t *)dst_YUV420_U;
	uint8_t *dst_v = (uint8_t *)dst_YUV420_V;
	int stride_y = dst_YUV420_Y_pitch;
	int stride_u = dst_YUV420_U_pitch;
	int stride_v = dst_YUV420_V_pitch;
	int16_t *src_MB = (int16_t *)src;

	for (n=0; n<4; n++)
	{
		for ( row=0; row<8; row++)
		{
			for(col=0; col<8; col++)
				dst_y[col] = (uint8_t)src_MB[col];
		
			src_MB += 8;
			dst_y += stride_y;
		}

		dst_y = dst_y - stride_y*8 + 8;
	}

	for ( row=0; row<8; row++)
	{
		for(col=0; col<8; col++)
			dst_v[col] = (uint8_t)src_MB[col];
	
		src_MB += 8;
		dst_v += stride_v;
	}

	for ( row=0; row<8; row++)
	{
		for(col=0; col<8; col++)
			dst_u[col] = (uint8_t)src_MB[col];
	
		src_MB += 8;
		dst_u += stride_u;
	}
}


void deinterleave_MBborder_YV12_NTSC(uint32_t *src, uint32_t *dst_YUV420_Y, int dst_YUV420_Y_pitch, 
						  uint32_t *dst_YUV420_U, int dst_YUV420_U_pitch, 
					      uint32_t *dst_YUV420_V, int dst_YUV420_V_pitch
					      )
{
	int n, row, col;
	uint8_t *dst_y = (uint8_t *)dst_YUV420_Y;
	uint8_t *dst_u = (uint8_t *)dst_YUV420_U;
	uint8_t *dst_v = (uint8_t *)dst_YUV420_V;
	int stride_y = dst_YUV420_Y_pitch;
	int stride_u = dst_YUV420_U_pitch;
	int stride_v = dst_YUV420_V_pitch;
	int16_t *src_MB = (int16_t *)src;

	for (n=0; n<4; n++)
	{
		for ( row=0; row<8; row++)
		{
			for(col=0; col<8; col++)
				dst_y[col] = (uint8_t)src_MB[col];
		
			src_MB += 8;
			dst_y += stride_y;
		}

		if (n==1) 
			dst_y = dst_y - 8;
		else
			dst_y = dst_y - stride_y*8 + 8;

	}

	for ( row=0; row<8; row++)
	{
		for(col=0; col<4; col++)
			dst_v[col] = (uint8_t)src_MB[col];
	
		for(col=0; col<4; col++)
			dst_v[col+8*stride_v] = (uint8_t)src_MB[col+4];

		src_MB += 8;
		dst_v += stride_v;
	}

	for ( row=0; row<8; row++)
	{
		for(col=0; col<4; col++)
			dst_u[col] = (uint8_t)src_MB[col];
	
		for(col=0; col<4; col++)
			dst_u[col+8*stride_u] = (uint8_t)src_MB[col+4];

		src_MB += 8;
		dst_u += stride_u;
	}
}


void YUV411_to_YUY2( uint8_t* src_YUV420_Y, int src_YUV420_Y_pitch, 
					 uint8_t* src_YUV420_U, int src_YUV420_U_pitch, 
					 uint8_t* src_YUV420_V, int src_YUV420_V_pitch,
					 uint8_t* dst_YUY2, int dst_YUY2_pitch, int width, int height )
{
	int row, col;
	int width_quarter = width/4;
	uint8_t *src_y = src_YUV420_Y;
	uint8_t *src_u = src_YUV420_U;
	uint8_t *src_v = src_YUV420_V;
	uint8_t *dst = dst_YUY2;

	for ( row=0; row<height; row++)
	{
		for( col=0; col<(width_quarter-1); col++ )
		{
			dst[col*8+0] = src_y[col*4+0];
			dst[col*8+1] = src_u[col  +0];
			dst[col*8+2] = src_y[col*4+1];
			dst[col*8+3] = src_v[col  +0];
			dst[col*8+4] = src_y[col*4+2];
			dst[col*8+5] = (uint8_t)(((int)src_u[col  +0] + (int)src_u[col  +1] + 1)>>1);
			dst[col*8+6] = src_y[col*4+3];
			dst[col*8+7] = (uint8_t)(((int)src_v[col  +0] + (int)src_v[col  +1] + 1)>>1);
		}
		dst[col*8+0] = src_y[col*4+0];
		dst[col*8+1] = src_u[col  +0];
		dst[col*8+2] = src_y[col*4+1];
		dst[col*8+3] = src_v[col  +0];
		dst[col*8+4] = src_y[col*4+2];
		dst[col*8+5] = src_u[col  +0];
		dst[col*8+6] = src_y[col*4+3];
		dst[col*8+7] = src_v[col  +0];

		dst += dst_YUY2_pitch;

		src_y += src_YUV420_Y_pitch;
		src_u += src_YUV420_U_pitch;
		src_v += src_YUV420_V_pitch;
	}
}

