/******************************************************************************/
/*                                                                            */
/*  Cedocida DV-Codec                                                         */
/*                                                                            */
/*  Copyright (C) 2005 Andreas Dittrich <dittrich@eit.uni-kl.de>              */
/*  Project start: 15.12.2003                                                 */
/*  Last change  : 08.05.2005                                                 */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/*  This program is free software; you can redistribute it and/or modify      */
/*  it under the terms of the GNU General Public License as published by      */
/*  the Free Software Foundation; either version 2 of the License, or         */
/*  (at your option) any later version.                                       */
/*                                                                            */
/*  This program is distributed in the hope that it will be useful,           */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/*  GNU General Public License for more details.                              */
/*                                                                            */
/*  You should have received a copy of the GNU General Public License         */
/*  along with this program; if not, write to the Free Software               */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
/*                                                                            */
/******************************************************************************/


#include "../inttypes.h"

#include "../signalproc.h"
#include "bitstream.h"

#include "vlc.h"
#include "encode_vlc.h"
#include "encode_vlc_tables.h"




#define VLC32DATA_SIZE 64*6*2
extern uint32_t MB_VLC32data[VLC32DATA_SIZE*5];	





static void block_scanning_88_unrolled( int16_t *block, block_bitbuffer *blk_bb, MB_VLC32buffer *MB_VLC32b );
static void block_scanning_248_unrolled( int16_t *block, block_bitbuffer *blk_bb, MB_VLC32buffer *MB_VLC32b );
static void put_vlc_code( block_bitbuffer *blk_bb, MB_VLC32buffer *MB_VLC32b, const uint32_t run, int32_t amp);

void block_scanning_88_unrolled_asm( int16_t *block, block_bitbuffer *blk_bb, MB_VLC32buffer *MB_VLC32b );
void block_scanning_248_unrolled_asm( int16_t *block, block_bitbuffer *blk_bb, MB_VLC32buffer *MB_VLC32b );



static uint32_t get_vlc_length( const uint32_t run, int32_t amp);
unsigned long get_vlc_length_asm(  unsigned long run, long amp );








static uint32_t get_vlc_length( const uint32_t run, int32_t amp)
{
	uint32_t length;

	if (amp<0) amp = -amp;
	if ((run)==0)
		length = (uint32_t)(vlc_table2[(amp)].length);
	else if (((amp)<=17) && ((run)<=4))
		length = vlc_table3[(run)*17+(amp)-18].length;
	else if (((amp)<=3) && ((run)<=14))
		length = vlc_table4[(run)*3+(amp)-16].length;
	else
	{
		length = (uint32_t)(vlc_table1[(run)-1].length);
		length = length + (uint32_t)(vlc_table2[(amp)].length);
	}
	return length;
}



#define GET_VLC_LENGTH_88_K( k ) \
	signed_amp = block[scan_order_88[k]];	\
	if (signed_amp!=0)	\
	{	\
		run = run + (k-1);	\
		length = length + get_vlc_length(run, signed_amp);	\
		run = -k;	\
	}

#define GET_VLC_LENGTH_248_K( k ) \
	signed_amp = block[scan_order_248[k]];	\
	if (signed_amp!=0)	\
	{	\
		run = run + (k-1);	\
		length = length + get_vlc_length( run, signed_amp);	\
		run = -k; 	\
	}

uint32_t get_vlc_length_88_unrolled( int16_t *block )
{
	uint32_t length = 0;
	int32_t run = 0;
	int32_t signed_amp;

	GET_VLC_LENGTH_88_K(  1 )
	GET_VLC_LENGTH_88_K(  2 ) 
	GET_VLC_LENGTH_88_K(  3 ) 
	GET_VLC_LENGTH_88_K(  4 ) 
	GET_VLC_LENGTH_88_K(  5 ) 
	GET_VLC_LENGTH_88_K(  6 ) 
	GET_VLC_LENGTH_88_K(  7 ) 
	GET_VLC_LENGTH_88_K(  8 ) 
	GET_VLC_LENGTH_88_K(  9 ) 
	GET_VLC_LENGTH_88_K( 10 ) 
	GET_VLC_LENGTH_88_K( 11 ) 
	GET_VLC_LENGTH_88_K( 12 ) 
	GET_VLC_LENGTH_88_K( 13 ) 
	GET_VLC_LENGTH_88_K( 14 ) 
	GET_VLC_LENGTH_88_K( 15 ) 
	GET_VLC_LENGTH_88_K( 16 ) 
	GET_VLC_LENGTH_88_K( 17 ) 
	GET_VLC_LENGTH_88_K( 18 ) 
	GET_VLC_LENGTH_88_K( 19 ) 
	GET_VLC_LENGTH_88_K( 20 ) 
	GET_VLC_LENGTH_88_K( 21 ) 
	GET_VLC_LENGTH_88_K( 22 ) 
	GET_VLC_LENGTH_88_K( 23 ) 
	GET_VLC_LENGTH_88_K( 24 ) 
	GET_VLC_LENGTH_88_K( 25 ) 
	GET_VLC_LENGTH_88_K( 26 ) 
	GET_VLC_LENGTH_88_K( 27 ) 
	GET_VLC_LENGTH_88_K( 28 ) 
	GET_VLC_LENGTH_88_K( 29 ) 
	GET_VLC_LENGTH_88_K( 30 ) 
	GET_VLC_LENGTH_88_K( 31 ) 
	GET_VLC_LENGTH_88_K( 32 ) 
	GET_VLC_LENGTH_88_K( 33 ) 
	GET_VLC_LENGTH_88_K( 34 ) 
	GET_VLC_LENGTH_88_K( 35 ) 
	GET_VLC_LENGTH_88_K( 36 ) 
	GET_VLC_LENGTH_88_K( 37 ) 
	GET_VLC_LENGTH_88_K( 38 ) 
	GET_VLC_LENGTH_88_K( 39 ) 
	GET_VLC_LENGTH_88_K( 40 ) 
	GET_VLC_LENGTH_88_K( 41 ) 
	GET_VLC_LENGTH_88_K( 42 ) 
	GET_VLC_LENGTH_88_K( 43 ) 
	GET_VLC_LENGTH_88_K( 44 ) 
	GET_VLC_LENGTH_88_K( 45 ) 
	GET_VLC_LENGTH_88_K( 46 ) 
	GET_VLC_LENGTH_88_K( 47 ) 
	GET_VLC_LENGTH_88_K( 48 ) 
	GET_VLC_LENGTH_88_K( 49 ) 
	GET_VLC_LENGTH_88_K( 50 ) 
	GET_VLC_LENGTH_88_K( 51 ) 
	GET_VLC_LENGTH_88_K( 52 ) 
	GET_VLC_LENGTH_88_K( 53 ) 
	GET_VLC_LENGTH_88_K( 54 ) 
	GET_VLC_LENGTH_88_K( 55 ) 
	GET_VLC_LENGTH_88_K( 56 ) 
	GET_VLC_LENGTH_88_K( 57 ) 
	GET_VLC_LENGTH_88_K( 58 ) 
	GET_VLC_LENGTH_88_K( 59 ) 
	GET_VLC_LENGTH_88_K( 60 ) 
	GET_VLC_LENGTH_88_K( 61 ) 
	GET_VLC_LENGTH_88_K( 62 ) 
	GET_VLC_LENGTH_88_K( 63 )	

	//if (run!=(-63)) length = length + 4;	//EOB
	length = length + 4;	//EOB

	return length;
}


uint32_t get_vlc_length_248_unrolled( int16_t *block )
{
	uint32_t length = 0;
	int32_t run = 0;
	int32_t signed_amp;

	GET_VLC_LENGTH_248_K(  1 )
	GET_VLC_LENGTH_248_K(  2 ) 
	GET_VLC_LENGTH_248_K(  3 ) 
	GET_VLC_LENGTH_248_K(  4 ) 
	GET_VLC_LENGTH_248_K(  5 ) 
	GET_VLC_LENGTH_248_K(  6 ) 
	GET_VLC_LENGTH_248_K(  7 ) 
	GET_VLC_LENGTH_248_K(  8 ) 
	GET_VLC_LENGTH_248_K(  9 ) 
	GET_VLC_LENGTH_248_K( 10 ) 
	GET_VLC_LENGTH_248_K( 11 ) 
	GET_VLC_LENGTH_248_K( 12 ) 
	GET_VLC_LENGTH_248_K( 13 ) 
	GET_VLC_LENGTH_248_K( 14 ) 
	GET_VLC_LENGTH_248_K( 15 ) 
	GET_VLC_LENGTH_248_K( 16 ) 
	GET_VLC_LENGTH_248_K( 17 ) 
	GET_VLC_LENGTH_248_K( 18 ) 
	GET_VLC_LENGTH_248_K( 19 ) 
	GET_VLC_LENGTH_248_K( 20 ) 
	GET_VLC_LENGTH_248_K( 21 ) 
	GET_VLC_LENGTH_248_K( 22 ) 
	GET_VLC_LENGTH_248_K( 23 ) 
	GET_VLC_LENGTH_248_K( 24 ) 
	GET_VLC_LENGTH_248_K( 25 ) 
	GET_VLC_LENGTH_248_K( 26 ) 
	GET_VLC_LENGTH_248_K( 27 ) 
	GET_VLC_LENGTH_248_K( 28 ) 
	GET_VLC_LENGTH_248_K( 29 ) 
	GET_VLC_LENGTH_248_K( 30 ) 
	GET_VLC_LENGTH_248_K( 31 ) 
	GET_VLC_LENGTH_248_K( 32 ) 
	GET_VLC_LENGTH_248_K( 33 ) 
	GET_VLC_LENGTH_248_K( 34 ) 
	GET_VLC_LENGTH_248_K( 35 ) 
	GET_VLC_LENGTH_248_K( 36 ) 
	GET_VLC_LENGTH_248_K( 37 ) 
	GET_VLC_LENGTH_248_K( 38 ) 
	GET_VLC_LENGTH_248_K( 39 ) 
	GET_VLC_LENGTH_248_K( 40 ) 
	GET_VLC_LENGTH_248_K( 41 ) 
	GET_VLC_LENGTH_248_K( 42 ) 
	GET_VLC_LENGTH_248_K( 43 ) 
	GET_VLC_LENGTH_248_K( 44 ) 
	GET_VLC_LENGTH_248_K( 45 ) 
	GET_VLC_LENGTH_248_K( 46 ) 
	GET_VLC_LENGTH_248_K( 47 ) 
	GET_VLC_LENGTH_248_K( 48 ) 
	GET_VLC_LENGTH_248_K( 49 ) 
	GET_VLC_LENGTH_248_K( 50 ) 
	GET_VLC_LENGTH_248_K( 51 ) 
	GET_VLC_LENGTH_248_K( 52 ) 
	GET_VLC_LENGTH_248_K( 53 ) 
	GET_VLC_LENGTH_248_K( 54 ) 
	GET_VLC_LENGTH_248_K( 55 ) 
	GET_VLC_LENGTH_248_K( 56 ) 
	GET_VLC_LENGTH_248_K( 57 ) 
	GET_VLC_LENGTH_248_K( 58 ) 
	GET_VLC_LENGTH_248_K( 59 ) 
	GET_VLC_LENGTH_248_K( 60 ) 
	GET_VLC_LENGTH_248_K( 61 ) 
	GET_VLC_LENGTH_248_K( 62 ) 
	GET_VLC_LENGTH_248_K( 63 )	

	//if (run!=(-63)) length = length + 4;	//EOB
	length = length + 4;	//EOB

	return length;
}


static void put_vlc_code( block_bitbuffer *blk_bb, MB_VLC32buffer *MB_VLC32b, const uint32_t run, int32_t amp)
{
	if (amp>=0)
	{
		if ((run)==0) 	
		{	
			uint32_t code = (uint32_t)(vlc_table2[(amp)].code);	
			uint32_t length = (uint32_t)(vlc_table2[(amp)].length);	
			putbits( blk_bb, MB_VLC32b, code, length );	
		}	
		else if (((amp)<=17) && ((run)<=4))	
		{	
			uint32_t code = vlc_table3[(run)*17+(amp)-18].code;	
			uint32_t length = vlc_table3[(run)*17+(amp)-18].length;	
			putbits( blk_bb, MB_VLC32b, code, length );	
		}	
		else if (((amp)<=3) && ((run)<=14))	
		{	
			uint32_t code = vlc_table4[(run)*3+(amp)-16].code;	
			uint32_t length = vlc_table4[(run)*3+(amp)-16].length;	
			putbits( blk_bb, MB_VLC32b, code, length );	
		}	
		else	
		{	
			uint32_t code1 = (uint32_t)(vlc_table1[(run)-1].code);	
			uint32_t length1 = (uint32_t)(vlc_table1[(run)-1].length);	
			uint32_t code = (uint32_t)(vlc_table2[(amp)].code);	
			uint32_t length = (uint32_t)(vlc_table2[(amp)].length);	
			code = ( (code1<<length) | code );	
			length = length+length1;	
			putbits( blk_bb, MB_VLC32b, code, length );	
		}	
	}
	else
	{
		amp = -amp;
		if ((run)==0) 	
		{	
			uint32_t code = (uint32_t)(vlc_table2[(amp)].code) + 1;	
			uint32_t length = (uint32_t)(vlc_table2[(amp)].length);	
			putbits( blk_bb, MB_VLC32b, code, length );	
		}	
		else if (((amp)<=17) && ((run)<=4))	
		{	
			uint32_t code = vlc_table3[(run)*17+(amp)-18].code + 1;	
			uint32_t length = vlc_table3[(run)*17+(amp)-18].length;	
			putbits( blk_bb, MB_VLC32b, code, length );	
		}	
		else if (((amp)<=3) && ((run)<=14))	
		{	
			uint32_t code = vlc_table4[(run)*3+(amp)-16].code + 1;	
			uint32_t length = vlc_table4[(run)*3+(amp)-16].length;	
			putbits( blk_bb, MB_VLC32b, code, length );	
		}	
		else	
		{	
			uint32_t code1 = (uint32_t)(vlc_table1[(run)-1].code);	
			uint32_t length1 = (uint32_t)(vlc_table1[(run)-1].length);	
			uint32_t code = (uint32_t)(vlc_table2[(amp)].code + 1);	
			uint32_t length = (uint32_t)(vlc_table2[(amp)].length);	
			code = ( (code1<<length) | code );	
			length = length+length1;	
			putbits( blk_bb, MB_VLC32b, code, length );	
		}
	}
}

#define BLOCK_SCANNING_88_K( k ) \
	signed_amp = (int32_t)( block[scan_order_88[k]] );	\
	if (signed_amp!=0)	\
	{	\
		run = run + (k-1);	\
		put_vlc_code( blk_bb, MB_VLC32b, run, signed_amp);	\
		run = -k;	\
	}

#define BLOCK_SCANNING_248_K( k ) \
	signed_amp = (int32_t)( block[scan_order_248[k]] );	\
	if (signed_amp!=0)	\
	{	\
		run = run + (k-1);	\
		put_vlc_code( blk_bb, MB_VLC32b, run, signed_amp);	\
		run = -k;	\
	}

static void block_scanning_88_unrolled( int16_t *block, block_bitbuffer *blk_bb, MB_VLC32buffer *MB_VLC32b )
{
	int32_t run = 0;
	int32_t signed_amp;

	BLOCK_SCANNING_88_K(  1 ) 
	BLOCK_SCANNING_88_K(  2 ) 
	BLOCK_SCANNING_88_K(  3 ) 
	BLOCK_SCANNING_88_K(  4 ) 
	BLOCK_SCANNING_88_K(  5 ) 
	BLOCK_SCANNING_88_K(  6 ) 
	BLOCK_SCANNING_88_K(  7 ) 
	BLOCK_SCANNING_88_K(  8 ) 
	BLOCK_SCANNING_88_K(  9 ) 
	BLOCK_SCANNING_88_K( 10 ) 
	BLOCK_SCANNING_88_K( 11 ) 
	BLOCK_SCANNING_88_K( 12 ) 
	BLOCK_SCANNING_88_K( 13 ) 
	BLOCK_SCANNING_88_K( 14 ) 
	BLOCK_SCANNING_88_K( 15 ) 
	BLOCK_SCANNING_88_K( 16 ) 
	BLOCK_SCANNING_88_K( 17 ) 
	BLOCK_SCANNING_88_K( 18 ) 
	BLOCK_SCANNING_88_K( 19 ) 
	BLOCK_SCANNING_88_K( 20 ) 
	BLOCK_SCANNING_88_K( 21 ) 
	BLOCK_SCANNING_88_K( 22 ) 
	BLOCK_SCANNING_88_K( 23 ) 
	BLOCK_SCANNING_88_K( 24 ) 
	BLOCK_SCANNING_88_K( 25 ) 
	BLOCK_SCANNING_88_K( 26 ) 
	BLOCK_SCANNING_88_K( 27 ) 
	BLOCK_SCANNING_88_K( 28 ) 
	BLOCK_SCANNING_88_K( 29 ) 
	BLOCK_SCANNING_88_K( 30 ) 
	BLOCK_SCANNING_88_K( 31 ) 
	BLOCK_SCANNING_88_K( 32 ) 
	BLOCK_SCANNING_88_K( 33 ) 
	BLOCK_SCANNING_88_K( 34 ) 
	BLOCK_SCANNING_88_K( 35 ) 
	BLOCK_SCANNING_88_K( 36 ) 
	BLOCK_SCANNING_88_K( 37 ) 
	BLOCK_SCANNING_88_K( 38 ) 
	BLOCK_SCANNING_88_K( 39 ) 
	BLOCK_SCANNING_88_K( 40 ) 
	BLOCK_SCANNING_88_K( 41 ) 
	BLOCK_SCANNING_88_K( 42 ) 
	BLOCK_SCANNING_88_K( 43 ) 
	BLOCK_SCANNING_88_K( 44 ) 
	BLOCK_SCANNING_88_K( 45 ) 
	BLOCK_SCANNING_88_K( 46 ) 
	BLOCK_SCANNING_88_K( 47 ) 
	BLOCK_SCANNING_88_K( 48 ) 
	BLOCK_SCANNING_88_K( 49 ) 
	BLOCK_SCANNING_88_K( 50 ) 
	BLOCK_SCANNING_88_K( 51 ) 
	BLOCK_SCANNING_88_K( 52 ) 
	BLOCK_SCANNING_88_K( 53 ) 
	BLOCK_SCANNING_88_K( 54 ) 
	BLOCK_SCANNING_88_K( 55 ) 
	BLOCK_SCANNING_88_K( 56 ) 
	BLOCK_SCANNING_88_K( 57 ) 
	BLOCK_SCANNING_88_K( 58 ) 
	BLOCK_SCANNING_88_K( 59 ) 
	BLOCK_SCANNING_88_K( 60 ) 
	BLOCK_SCANNING_88_K( 61 ) 
	BLOCK_SCANNING_88_K( 62 ) 
	BLOCK_SCANNING_88_K( 63 )	

	//if (run!=(-63)) putbits( blk_bb, MB_VLC32b, 0x06, 4 );	//EOB
	putbits( blk_bb, MB_VLC32b, 0x06, 4 );	//EOB
}


static void block_scanning_248_unrolled( int16_t *block, block_bitbuffer *blk_bb, MB_VLC32buffer *MB_VLC32b )
{
	int32_t run = 0;
	int32_t signed_amp;

	BLOCK_SCANNING_248_K(  1 ) 
	BLOCK_SCANNING_248_K(  2 ) 
	BLOCK_SCANNING_248_K(  3 ) 
	BLOCK_SCANNING_248_K(  4 ) 
	BLOCK_SCANNING_248_K(  5 ) 
	BLOCK_SCANNING_248_K(  6 ) 
	BLOCK_SCANNING_248_K(  7 ) 
	BLOCK_SCANNING_248_K(  8 ) 
	BLOCK_SCANNING_248_K(  9 ) 
	BLOCK_SCANNING_248_K( 10 ) 
	BLOCK_SCANNING_248_K( 11 ) 
	BLOCK_SCANNING_248_K( 12 ) 
	BLOCK_SCANNING_248_K( 13 ) 
	BLOCK_SCANNING_248_K( 14 ) 
	BLOCK_SCANNING_248_K( 15 ) 
	BLOCK_SCANNING_248_K( 16 ) 
	BLOCK_SCANNING_248_K( 17 ) 
	BLOCK_SCANNING_248_K( 18 ) 
	BLOCK_SCANNING_248_K( 19 ) 
	BLOCK_SCANNING_248_K( 20 ) 
	BLOCK_SCANNING_248_K( 21 ) 
	BLOCK_SCANNING_248_K( 22 ) 
	BLOCK_SCANNING_248_K( 23 ) 
	BLOCK_SCANNING_248_K( 24 ) 
	BLOCK_SCANNING_248_K( 25 ) 
	BLOCK_SCANNING_248_K( 26 ) 
	BLOCK_SCANNING_248_K( 27 ) 
	BLOCK_SCANNING_248_K( 28 ) 
	BLOCK_SCANNING_248_K( 29 ) 
	BLOCK_SCANNING_248_K( 30 ) 
	BLOCK_SCANNING_248_K( 31 ) 
	BLOCK_SCANNING_248_K( 32 ) 
	BLOCK_SCANNING_248_K( 33 ) 
	BLOCK_SCANNING_248_K( 34 ) 
	BLOCK_SCANNING_248_K( 35 ) 
	BLOCK_SCANNING_248_K( 36 ) 
	BLOCK_SCANNING_248_K( 37 ) 
	BLOCK_SCANNING_248_K( 38 ) 
	BLOCK_SCANNING_248_K( 39 ) 
	BLOCK_SCANNING_248_K( 40 ) 
	BLOCK_SCANNING_248_K( 41 ) 
	BLOCK_SCANNING_248_K( 42 ) 
	BLOCK_SCANNING_248_K( 43 ) 
	BLOCK_SCANNING_248_K( 44 ) 
	BLOCK_SCANNING_248_K( 45 ) 
	BLOCK_SCANNING_248_K( 46 ) 
	BLOCK_SCANNING_248_K( 47 ) 
	BLOCK_SCANNING_248_K( 48 ) 
	BLOCK_SCANNING_248_K( 49 ) 
	BLOCK_SCANNING_248_K( 50 ) 
	BLOCK_SCANNING_248_K( 51 ) 
	BLOCK_SCANNING_248_K( 52 ) 
	BLOCK_SCANNING_248_K( 53 ) 
	BLOCK_SCANNING_248_K( 54 ) 
	BLOCK_SCANNING_248_K( 55 ) 
	BLOCK_SCANNING_248_K( 56 ) 
	BLOCK_SCANNING_248_K( 57 ) 
	BLOCK_SCANNING_248_K( 58 ) 
	BLOCK_SCANNING_248_K( 59 ) 
	BLOCK_SCANNING_248_K( 60 ) 
	BLOCK_SCANNING_248_K( 61 ) 
	BLOCK_SCANNING_248_K( 62 ) 
	BLOCK_SCANNING_248_K( 63 )

	//if (run!=(-63)) putbits( blk_bb, MB_VLC32b, 0x06, 4 );	//EOB
	putbits( blk_bb, MB_VLC32b, 0x06, 4 );	//EOB
}


void __declspec(naked) block_scanning_88_unrolled_asm( int16_t *block, block_bitbuffer *blk_bb, MB_VLC32buffer *MB_VLC32b )
{
	__asm {
		
		push ebx
		push ecx
		push esi
		push edi

		mov ebx, DWORD PTR [esp+ 4+(4*4)]	; block
		mov esi, DWORD PTR [esp+ 8+(4*4)]	; blk_bb
		mov edi, DWORD PTR [esp+ 12+(4*4)]	; MB_VLC32b

		sub esp, 8	; placeholder *amp* / *length*,  placeholder *run* / *code*
		push edi	; MB_VLC32b
		push esi	; blk_bb

		xor eax, eax	; run
		movsx	esi, WORD PTR [ebx+2]

		test esi, esi
		movsx edi, WORD PTR [ebx+16]
		je c02
		mov [esp+12],esi
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -1
	c02:
		test edi, edi
		movsx esi, WORD PTR [ebx+32]
		je c03
		mov [esp+12],edi
		add eax, 1
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -2
	c03:
		test esi, esi
		movsx edi, WORD PTR [ebx+18]
		je c04
		mov [esp+12],esi
		add eax, 2
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -3
	c04:
		test edi, edi
		movsx esi, WORD PTR [ebx+ 4]
		je c05
		mov [esp+12],edi
		add eax, 3
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -4
	c05:
		test esi, esi
		movsx edi, WORD PTR [ebx+ 6]
		je c06
		mov [esp+12],esi
		add eax, 4
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -5
	c06:
		test edi, edi
		movsx esi, WORD PTR [ebx+20]
		je c07
		mov [esp+12],edi
		add eax, 5
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -6
	c07:
		test esi, esi
		movsx edi, WORD PTR [ebx+34]
		je c08
		mov [esp+12],esi
		add eax, 6
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -7
	c08:
		test edi, edi
		movsx esi, WORD PTR [ebx+48]
		je c09
		mov [esp+12],edi
		add eax, 7
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -8
	c09:
		test esi, esi
		movsx edi, WORD PTR [ebx+64]
		je c10
		mov [esp+12],esi
		add eax, 8
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -9
	c10:
		test edi, edi
		movsx esi, WORD PTR [ebx+50]
		je c11
		mov [esp+12],edi
		add eax, 9
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -10
	c11:
		test esi, esi
		movsx edi, WORD PTR [ebx+36]
		je c12
		mov [esp+12],esi
		add eax, 10
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -11
	c12:
		test edi, edi
		movsx esi, WORD PTR [ebx+22]
		je c13
		mov [esp+12],edi
		add eax, 11
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -12
	c13:
		test esi, esi
		movsx edi, WORD PTR [ebx+ 8]
		je c14
		mov [esp+12],esi
		add eax, 12
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -13
	c14:
		test edi, edi
		movsx esi, WORD PTR [ebx+10]
		je c15
		mov [esp+12],edi
		add eax, 13
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -14
	c15:
		test esi, esi
		movsx edi, WORD PTR [ebx+24]
		je c16
		mov [esp+12],esi
		add eax, 14
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -15
	c16:
		test edi, edi
		movsx esi, WORD PTR [ebx+38]
		je c17
		mov [esp+12],edi
		add eax, 15
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -16
	c17:
		test esi, esi
		movsx edi, WORD PTR [ebx+52]
		je c18
		mov [esp+12],esi
		add eax, 16
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -17
	c18:
		test edi, edi
		movsx esi, WORD PTR [ebx+66]
		je c19
		mov [esp+12],edi
		add eax, 17
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -18
	c19:
		test esi, esi
		movsx edi, WORD PTR [ebx+80]
		je c20
		mov [esp+12],esi
		add eax, 18
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -19
	c20:
		test edi, edi
		movsx esi, WORD PTR [ebx+96]
		je c21
		mov [esp+12],edi
		add eax, 19
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -20
	c21:
		test esi, esi
		movsx edi, WORD PTR [ebx+82]
		je c22
		mov [esp+12],esi
		add eax, 20
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -21
	c22:
		test edi, edi
		movsx esi, WORD PTR [ebx+68]
		je c23
		mov [esp+12],edi
		add eax, 21
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -22
	c23:
		test esi, esi
		movsx edi, WORD PTR [ebx+54]
		je c24
		mov [esp+12],esi
		add eax, 22
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -23
	c24:
		test edi, edi
		movsx esi, WORD PTR [ebx+40]
		je c25
		mov [esp+12],edi
		add eax, 23
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -24
	c25:
		test esi, esi
		movsx edi, WORD PTR [ebx+26]
		je c26
		mov [esp+12],esi
		add eax, 24
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -25
	c26:
		test edi, edi
		movsx esi, WORD PTR [ebx+12]
		je c27
		mov [esp+12],edi
		add eax, 25
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -26
	c27:
		test esi, esi
		movsx edi, WORD PTR [ebx+14]
		je c28
		mov [esp+12],esi
		add eax, 26
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -27
	c28:
		test edi, edi
		movsx esi, WORD PTR [ebx+28]
		je c29
		mov [esp+12],edi
		add eax, 27
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -28
	c29:
		test esi, esi
		movsx edi, WORD PTR [ebx+42]
		je c30
		mov [esp+12],esi
		add eax, 28
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -29
	c30:
		test edi, edi
		movsx esi, WORD PTR [ebx+56]
		je c31
		mov [esp+12],edi
		add eax, 29
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -30
	c31:
		test esi, esi
		movsx edi, WORD PTR [ebx+70]
		je c32
		mov [esp+12],esi
		add eax, 30
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -31
	c32:
		test edi, edi
		movsx esi, WORD PTR [ebx+84]
		je c33
		mov [esp+12],edi
		add eax, 31
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -32
	c33:
		test esi, esi
		movsx edi, WORD PTR [ebx+98]
		je c34
		mov [esp+12],esi
		add eax, 32
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -33
	c34:
		test edi, edi
		movsx esi, WORD PTR [ebx+112]
		je c35
		mov [esp+12],edi
		add eax, 33
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -34
	c35:
		test esi, esi
		movsx edi, WORD PTR [ebx+114]
		je c36
		mov [esp+12],esi
		add eax, 34
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -35
	c36:
		test edi, edi
		movsx esi, WORD PTR [ebx+100]
		je c37
		mov [esp+12],edi
		add eax, 35
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -36
	c37:
		test esi, esi
		movsx edi, WORD PTR [ebx+86]
		je c38
		mov [esp+12],esi
		add eax, 36
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -37
	c38:
		test edi, edi
		movsx esi, WORD PTR [ebx+72]
		je c39
		mov [esp+12],edi
		add eax, 37
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -38
	c39:
		test esi, esi
		movsx edi, WORD PTR [ebx+58]
		je c40
		mov [esp+12],esi
		add eax, 38
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -39
	c40:
		test edi, edi
		movsx esi, WORD PTR [ebx+44]
		je c41
		mov [esp+12],edi
		add eax, 39
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -40
	c41:
		test esi, esi
		movsx edi, WORD PTR [ebx+30]
		je c42
		mov [esp+12],esi
		add eax, 40
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -41
	c42:
		test edi, edi
		movsx esi, WORD PTR [ebx+46]
		je c43
		mov [esp+12],edi
		add eax, 41
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -42
	c43:
		test esi, esi
		movsx edi, WORD PTR [ebx+60]
		je c44
		mov [esp+12],esi
		add eax, 42
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -43
	c44:
		test edi, edi
		movsx esi, WORD PTR [ebx+74]
		je c45
		mov [esp+12],edi
		add eax, 43
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -44
	c45:
		test esi, esi
		movsx edi, WORD PTR [ebx+88]
		je c46
		mov [esp+12],esi
		add eax, 44
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -45
	c46:
		test edi, edi
		movsx esi, WORD PTR [ebx+102]
		je c47
		mov [esp+12],edi
		add eax, 45
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -46
	c47:

		mov edi, DWORD PTR [ebx+76]		;76 & 78
		mov ecx, esi
		mov esi, DWORD PTR [ebx+92]		; 92 & 94
		or ecx, edi
		mov edi, DWORD PTR [ebx+104]	;104 & 106
		mov ecx, esi
		mov esi, DWORD PTR [ebx+108]	;108 & 110
		or ecx, edi
		mov edi, DWORD PTR [ebx+116]	;116 & 118
		mov ecx, esi
		mov esi, DWORD PTR [ebx+120]	;120 & 122
		or ecx, edi
		mov edi, DWORD PTR [ebx+124]	;124 & 126
		mov ecx, esi
		movsx esi, WORD PTR [ebx+62]
		or ecx, edi
		movsx edi, WORD PTR [ebx+90]
		mov ecx, esi
		or ecx, edi
		je c64
		movsx esi, WORD PTR [ebx+102]


		test esi, esi
		movsx edi, WORD PTR [ebx+116]
		je c48
		mov [esp+12],esi
		add eax, 46
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -47
	c48:
		test edi, edi
		movsx esi, WORD PTR [ebx+118]
		je c49
		mov [esp+12],edi
		add eax, 47
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -48
	c49:
		test esi, esi
		movsx edi, WORD PTR [ebx+104]
		je c50
		mov [esp+12],esi
		add eax, 48
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -49
	c50:
		test edi, edi
		movsx esi, WORD PTR [ebx+90]
		je c51
		mov [esp+12],edi
		add eax, 49
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -50
	c51:
		test esi, esi
		movsx edi, WORD PTR [ebx+76]
		je c52
		mov [esp+12],esi
		add eax, 50
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -51
	c52:
		test edi, edi
		movsx esi, WORD PTR [ebx+62]
		je c53
		mov [esp+12],edi
		add eax, 51
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -52
	c53:
		test esi, esi
		movsx edi, WORD PTR [ebx+78]
		je c54
		mov [esp+12],esi
		add eax, 52
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -53
	c54:
		test edi, edi
		movsx esi, WORD PTR [ebx+92]
		je c55
		mov [esp+12],edi
		add eax, 53
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -54
	c55:
		test esi, esi
		movsx edi, WORD PTR [ebx+106]
		je c56
		mov [esp+12],esi
		add eax, 54
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -55
	c56:
		test edi, edi
		movsx esi, WORD PTR [ebx+120]
		je c57
		mov [esp+12],edi
		add eax, 55
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -56
	c57:
		test esi, esi
		movsx edi, WORD PTR [ebx+122]
		je c58
		mov [esp+12],esi
		add eax, 56
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -57
	c58:
		test edi, edi
		movsx esi, WORD PTR [ebx+108]
		je c59
		mov [esp+12],edi
		add eax, 57
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -58
	c59:
		test esi, esi
		movsx edi, WORD PTR [ebx+94]
		je c60
		mov [esp+12],esi
		add eax, 58
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -59
	c60:
		test edi, edi
		movsx esi, WORD PTR [ebx+110]
		je c61
		mov [esp+12],edi
		add eax, 59
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -60
	c61:
		test esi, esi
		movsx edi, WORD PTR [ebx+124]
		je c62
		mov [esp+12],esi
		add eax, 60
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -61
	c62:
		test edi, edi
		movsx esi, WORD PTR [ebx+126]
		je c63
		mov [esp+12],edi
		add eax, 61
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -62

	c63:
		test esi, esi
		je c64
		mov [esp+12],esi
		add eax, 62
		mov [esp+8], eax
		call put_vlc_code
		//jmp block_scanning_end
	c64:
		mov DWORD PTR [esp+8], 0x06
		mov DWORD PTR [esp+12],0x04
		call putbits
	//block_scanning_end:

/*
		rdtsc
		mov ebx, eax
	my_counter_loop:
		rdtsc
		sub eax, ebx
		sub eax, (64*5)
		jl my_counter_loop
*/

		add esp, 16
		pop edi;
		pop esi;
		pop ecx
		pop ebx;
		ret 0
	}
}


void __declspec(naked) block_scanning_248_unrolled_asm( int16_t *block, block_bitbuffer *blk_bb, MB_VLC32buffer *MB_VLC32b )
{
	__asm {
		
		push ebx
		push ecx
		push esi
		push edi

		mov ebx, DWORD PTR [esp+ 4+(4*4)]	; block
		mov esi, DWORD PTR [esp+ 8+(4*4)]	; blk_bb
		mov edi, DWORD PTR [esp+ 12+(4*4)]	; MB_VLC32b

		sub esp, 8	; placeholder *amp* / *length*,  placeholder *run* / *code*
		push edi	; MB_VLC32b
		push esi	; blk_bb

		xor eax, eax	; run
		movsx	esi, WORD PTR [ebx+64]

		test esi, esi
		movsx edi, WORD PTR [ebx+ 2]
		je c02
		mov [esp+12],esi
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -1

	c02:
		test edi, edi
		movsx esi, WORD PTR [ebx+66]
		je c03
		mov [esp+12],edi
		add eax, 1
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -2
	c03:
		test esi, esi
		movsx edi, WORD PTR [ebx+16]
		je c04
		mov [esp+12],esi
		add eax, 2
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -3
	c04:
		test edi, edi
		movsx esi, WORD PTR [ebx+80]
		je c05
		mov [esp+12],edi
		add eax, 3
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -4
	c05:
		test esi, esi
		movsx edi, WORD PTR [ebx+ 4]
		je c06
		mov [esp+12],esi
		add eax, 4
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -5
	c06:
		test edi, edi
		movsx esi, WORD PTR [ebx+68]
		je c07
		mov [esp+12],edi
		add eax, 5
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -6
	c07:
		test esi, esi
		movsx edi, WORD PTR [ebx+18]
		je c08
		mov [esp+12],esi
		add eax, 6
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -7
	c08:
		test edi, edi
		movsx esi, WORD PTR [ebx+82]
		je c09
		mov [esp+12],edi
		add eax, 7
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -8
	c09:
		test esi, esi
		movsx edi, WORD PTR [ebx+32]
		je c10
		mov [esp+12],esi
		add eax, 8
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -9
	c10:
		test edi, edi
		movsx esi, WORD PTR [ebx+96]
		je c11
		mov [esp+12],edi
		add eax, 9
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -10
	c11:
		test esi, esi
		movsx edi, WORD PTR [ebx+48]
		je c12
		mov [esp+12],esi
		add eax, 10
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -11
	c12:
		test edi, edi
		movsx esi, WORD PTR [ebx+112]
		je c13
		mov [esp+12],edi
		add eax, 11
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -12
	c13:
		test esi, esi
		movsx edi, WORD PTR [ebx+34]
		je c14
		mov [esp+12],esi
		add eax, 12
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -13
	c14:
		test edi, edi
		movsx esi, WORD PTR [ebx+98]
		je c15
		mov [esp+12],edi
		add eax, 13
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -14
	c15:
		test esi, esi
		movsx edi, WORD PTR [ebx+20]
		je c16
		mov [esp+12],esi
		add eax, 14
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -15
	c16:
		test edi, edi
		movsx esi, WORD PTR [ebx+84]
		je c17
		mov [esp+12],edi
		add eax, 15
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -16
	c17:
		test esi, esi
		movsx edi, WORD PTR [ebx+ 6]
		je c18
		mov [esp+12],esi
		add eax, 16
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -17
	c18:
		test edi, edi
		movsx esi, WORD PTR [ebx+70]
		je c19
		mov [esp+12],edi
		add eax, 17
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -18
	c19:
		test esi, esi
		movsx edi, WORD PTR [ebx+ 8]
		je c20
		mov [esp+12],esi
		add eax, 18
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -19
	c20:
		test edi, edi
		movsx esi, WORD PTR [ebx+72]
		je c21
		mov [esp+12],edi
		add eax, 19
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -20
	c21:
		test esi, esi
		movsx edi, WORD PTR [ebx+22]
		je c22
		mov [esp+12],esi
		add eax, 20
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -21
	c22:
		test edi, edi
		movsx esi, WORD PTR [ebx+86]
		je c23
		mov [esp+12],edi
		add eax, 21
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -22
	c23:
		test esi, esi
		movsx edi, WORD PTR [ebx+36]
		je c24
		mov [esp+12],esi
		add eax, 22
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -23
	c24:
		test edi, edi
		movsx esi, WORD PTR [ebx+100]
		je c25
		mov [esp+12],edi
		add eax, 23
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -24
	c25:
		test esi, esi
		movsx edi, WORD PTR [ebx+50]
		je c26
		mov [esp+12],esi
		add eax, 24
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -25
	c26:
		test edi, edi
		movsx esi, WORD PTR [ebx+114]
		je c27
		mov [esp+12],edi
		add eax, 25
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -26
	c27:
		test esi, esi
		movsx edi, WORD PTR [ebx+52]
		je c28
		mov [esp+12],esi
		add eax, 26
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -27
	c28:
		test edi, edi
		movsx esi, WORD PTR [ebx+116]
		je c29
		mov [esp+12],edi
		add eax, 27
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -28
	c29:
		test esi, esi
		movsx edi, WORD PTR [ebx+38]
		je c30
		mov [esp+12],esi
		add eax, 28
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -29
	c30:
		test edi, edi
		movsx esi, WORD PTR [ebx+102]
		je c31
		mov [esp+12],edi
		add eax, 29
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -30
	c31:
		test esi, esi
		movsx edi, WORD PTR [ebx+24]
		je c32
		mov [esp+12],esi
		add eax, 30
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -31
	c32:
		test edi, edi
		movsx esi, WORD PTR [ebx+88]
		je c33
		mov [esp+12],edi
		add eax, 31
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -32
	c33:
		test esi, esi
		movsx edi, WORD PTR [ebx+10]
		je c34
		mov [esp+12],esi
		add eax, 32
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -33
	c34:
		test edi, edi
		movsx esi, WORD PTR [ebx+74]
		je c35
		mov [esp+12],edi
		add eax, 33
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -34
	c35:
		test esi, esi
		movsx edi, WORD PTR [ebx+12]
		je c36
		mov [esp+12],esi
		add eax, 34
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -35
	c36:
		test edi, edi
		movsx esi, WORD PTR [ebx+76]
		je c37
		mov [esp+12],edi
		add eax, 35
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -36
	c37:
		test esi, esi
		movsx edi, WORD PTR [ebx+26]
		je c38
		mov [esp+12],esi
		add eax, 36
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -37
	c38:
		test edi, edi
		movsx esi, WORD PTR [ebx+90]
		je c39
		mov [esp+12],edi
		add eax, 37
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -38
	c39:
		test esi, esi
		movsx edi, WORD PTR [ebx+40]
		je c40
		mov [esp+12],esi
		add eax, 38
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -39
	c40:
		test edi, edi
		movsx esi, WORD PTR [ebx+104]
		je c41
		mov [esp+12],edi
		add eax, 39
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -40
	c41:
		test esi, esi
		movsx edi, WORD PTR [ebx+54]
		je c42
		mov [esp+12],esi
		add eax, 40
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -41
	c42:
		test edi, edi
		movsx esi, WORD PTR [ebx+118]
		je c43
		mov [esp+12],edi
		add eax, 41
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -42
	c43:
		test esi, esi
		movsx edi, WORD PTR [ebx+56]
		je c44
		mov [esp+12],esi
		add eax, 42
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -43
	c44:
		test edi, edi
		movsx esi, WORD PTR [ebx+120]
		je c45
		mov [esp+12],edi
		add eax, 43
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -44
	c45:
		test esi, esi
		movsx edi, WORD PTR [ebx+42]
		je c46
		mov [esp+12],esi
		add eax, 44
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -45
	c46:
		test edi, edi
		movsx esi, WORD PTR [ebx+106]
		je c47
		mov [esp+12],edi
		add eax, 45
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -46
	c47:
		test esi, esi
		movsx edi, WORD PTR [ebx+28]
		je c48
		mov [esp+12],esi
		add eax, 46
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -47
	c48:
		test edi, edi
		movsx esi, WORD PTR [ebx+92]
		je c49
		mov [esp+12],edi
		add eax, 47
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -48
	c49:
		test esi, esi
		movsx edi, WORD PTR [ebx+14]
		je c50
		mov [esp+12],esi
		add eax, 48
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -49
	c50:
		test edi, edi
		movsx esi, WORD PTR [ebx+78]
		je c51
		mov [esp+12],edi
		add eax, 49
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -50
	c51:
		test esi, esi
		movsx edi, WORD PTR [ebx+30]
		je c52
		mov [esp+12],esi
		add eax, 50
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -51
	c52:
		test edi, edi
		movsx esi, WORD PTR [ebx+94]
		je c53
		mov [esp+12],edi
		add eax, 51
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -52
	c53:
		test esi, esi
		movsx edi, WORD PTR [ebx+44]
		je c54
		mov [esp+12],esi
		add eax, 52
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -53
	c54:
		test edi, edi
		movsx esi, WORD PTR [ebx+108]
		je c55
		mov [esp+12],edi
		add eax, 53
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -54
	c55:
		test esi, esi
		movsx edi, WORD PTR [ebx+58]
		je c56
		mov [esp+12],esi
		add eax, 54
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -55
	c56:
		test edi, edi
		movsx esi, WORD PTR [ebx+122]
		je c57
		mov [esp+12],edi
		add eax, 55
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -56
	c57:
		test esi, esi
		movsx edi, WORD PTR [ebx+60]
		je c58
		mov [esp+12],esi
		add eax, 56
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -57
	c58:
		test edi, edi
		movsx esi, WORD PTR [ebx+124]
		je c59
		mov [esp+12],edi
		add eax, 57
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -58
	c59:
		test esi, esi
		movsx edi, WORD PTR [ebx+46]
		je c60
		mov [esp+12],esi
		add eax, 58
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -59
	c60:
		test edi, edi
		movsx esi, WORD PTR [ebx+110]
		je c61
		mov [esp+12],edi
		add eax, 59
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -60
	c61:
		test esi, esi
		movsx edi, WORD PTR [ebx+62]
		je c62
		mov [esp+12],esi
		add eax, 60
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -61
	c62:
		test edi, edi
		movsx esi, WORD PTR [ebx+126]
		je c63
		mov [esp+12],edi
		add eax, 61
		mov [esp+8], eax
		call put_vlc_code
		mov eax, -62

	c63:
		test esi, esi
		je c64
		mov [esp+12],esi
		add eax, 62
		mov [esp+8], eax
		call put_vlc_code
		//jmp block_scanning_end
	c64:
		mov DWORD PTR [esp+8], 0x06
		mov DWORD PTR [esp+12],0x04
		call putbits
	//block_scanning_end:

/*
		rdtsc
		mov ebx, eax
	my_counter_loop:
		rdtsc
		sub eax, ebx
		sub eax, (64*5)
		jl my_counter_loop
*/

		add esp, 16
		pop edi;
		pop esi;
		pop ecx
		pop ebx;
		ret 0
	}
}


unsigned long __declspec(naked) get_vlc_length_asm(  unsigned long run, long amp ) 
{
	__asm {
		
		push ebx
		push ecx

		mov ecx, DWORD PTR [esp+ 4+(4*2)]	; run
		mov ebx, DWORD PTR [esp+ 8+(4*2)]	; amp

		test ebx, ebx
		jns is_positive
	//is_negative:
		neg ebx
	is_positive:
		test ecx, ecx
		jne run_is_not_zero
	//run_is_zero:
		movzx eax, WORD PTR [ebx*4+(vlc_table2+2)]
		pop ecx
		pop ebx
		ret

	run_is_not_zero:
		sub ecx, 4
		jg run_is_greater_than_4

		sub ebx, 17
		jg amp_is_greater_than_17

		imul ecx, (17*8)
		mov eax, DWORD PTR [ecx+ebx*8+(16*8)+(17*8*3)+(vlc_table3+4)]
		pop ecx
		pop ebx
		ret

	run_is_greater_than_4:
		sub ecx, (14-4)
		jg run_is_greater_than_14

		sub ebx, 3
		jg amp_is_greater_than_3_run_is_greater_than_4

		imul ecx, (3*8)
		mov eax, DWORD PTR [ecx+(3*8*9)+ebx*8+(2*8)+vlc_table4+4]
		pop ecx
		pop ebx
		ret

	run_is_greater_than_14:
		movzx eax, WORD PTR [ecx*4+(14-1)*4+vlc_table1+2]
		movzx ecx, WORD PTR [ebx*4+vlc_table2+2]
		add eax, ecx
		pop ecx
		pop ebx
		ret

	amp_is_greater_than_17:
		movzx eax, WORD PTR [ecx*4+(4-1)*4+(-1+1)*4+vlc_table1+2]
		movzx ecx, WORD PTR [ebx*4+(17)*4+vlc_table2+2]
		add eax, ecx
		pop ecx
		pop ebx
		ret

	amp_is_greater_than_3_run_is_greater_than_4:
		movzx eax, WORD PTR [ecx*4+9*4+(4-1+1)*4+vlc_table1+2]
		movzx ecx, WORD PTR [ebx*4+(3)*4+vlc_table2+2]
		add eax, ecx
		pop ecx
		pop ebx
		ret

	}
}


unsigned long __declspec(naked) get_vlc_length_88_unrolled_asm( short *block )
{
	__asm {
		
		push ebx
		push ecx
		push edx
		push esi
		push edi

		mov ebx, DWORD PTR [esp+ 4+(4*5)]	; block
		xor ecx, ecx	; sum of length

		xor eax, eax	; run
		movsx	esi, WORD PTR [ebx+2]

		test esi, esi
		movsx edi, WORD PTR [ebx+16]
		je c02
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -1

	c02:
		test edi, edi
		movsx esi, WORD PTR [ebx+32]
		je c03
		add eax, 1
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -2
	c03:
		test esi, esi
		movsx edi, DWORD PTR [ebx+18]
		je c04
		add eax, 2
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -3
	c04:
		test edi, edi
		movsx esi, WORD PTR [ebx+ 4]
		je c05
		add eax, 3
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -4
	c05:
		test esi, esi
		movsx edi, DWORD PTR [ebx+ 6]
		je c06
		add eax, 4
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -5
	c06:
		test edi, edi
		movsx esi, WORD PTR [ebx+20]
		je c07
		add eax, 5
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -6
	c07:
		test esi, esi
		movsx edi, DWORD PTR [ebx+34]
		je c08
		add eax, 6
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -7
	c08:
		test edi, edi
		movsx esi, WORD PTR [ebx+48]
		je c09
		add eax, 7
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -8
	c09:
		test esi, esi
		movsx edi, DWORD PTR [ebx+64]
		je c10
		add eax, 8
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -9
	c10:
		test edi, edi
		movsx esi, WORD PTR [ebx+50]
		je c11
		add eax, 9
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -10
	c11:
		test esi, esi
		movsx edi, DWORD PTR [ebx+36]
		je c12
		add eax, 10
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -11
	c12:
		test edi, edi
		movsx esi, WORD PTR [ebx+22]
		je c13
		add eax, 11
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -12
	c13:
		test esi, esi
		movsx edi, DWORD PTR [ebx+ 8]
		je c14
		add eax, 12
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -13
	c14:
		test edi, edi
		movsx esi, WORD PTR [ebx+10]
		je c15
		add eax, 13
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -14
	c15:
		test esi, esi
		movsx edi, DWORD PTR [ebx+24]
		je c16
		add eax, 14
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -15
	c16:
		test edi, edi
		movsx esi, WORD PTR [ebx+38]
		je c17
		add eax, 15
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -16
	c17:
		test esi, esi
		movsx edi, DWORD PTR [ebx+52]
		je c18
		add eax, 16
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -17
	c18:
		test edi, edi
		movsx esi, WORD PTR [ebx+66]
		je c19
		add eax, 17
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -18
	c19:
		test esi, esi
		movsx edi, DWORD PTR [ebx+80]
		je c20
		add eax, 18
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -19
	c20:
		test edi, edi
		movsx esi, WORD PTR [ebx+96]
		je c21
		add eax, 19
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -20
	c21:
		test esi, esi
		movsx edi, DWORD PTR [ebx+82]
		je c22
		add eax, 20
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -21
	c22:
		test edi, edi
		movsx esi, WORD PTR [ebx+68]
		je c23
		add eax, 21
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -22
	c23:
		test esi, esi
		movsx edi, DWORD PTR [ebx+54]
		je c24
		add eax, 22
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -23
	c24:
		test edi, edi
		movsx esi, WORD PTR [ebx+40]
		je c25
		add eax, 23
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -24
	c25:
		test esi, esi
		movsx edi, DWORD PTR [ebx+26]
		je c26
		add eax, 24
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -25
	c26:
		test edi, edi
		movsx esi, WORD PTR [ebx+12]
		je c27
		add eax, 25
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -26
	c27:
		test esi, esi
		movsx edi, DWORD PTR [ebx+14]
		je c28
		add eax, 26
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -27
	c28:
		test edi, edi
		movsx esi, WORD PTR [ebx+28]
		je c29
		add eax, 27
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -28
	c29:
		test esi, esi
		movsx edi, DWORD PTR [ebx+42]
		je c30
		add eax, 28
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -29
	c30:
		test edi, edi
		movsx esi, WORD PTR [ebx+56]
		je c31
		add eax, 29
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -30
	c31:
		test esi, esi
		movsx edi, DWORD PTR [ebx+70]
		je c32
		add eax, 30
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -31
	c32:
		test edi, edi
		movsx esi, WORD PTR [ebx+84]
		je c33
		add eax, 31
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -32
	c33:
		test esi, esi
		movsx edi, DWORD PTR [ebx+98]
		je c34
		add eax, 32
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -33
	c34:
		test edi, edi
		movsx esi, WORD PTR [ebx+112]
		je c35
		add eax, 33
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -34
	c35:
		test esi, esi
		movsx edi, DWORD PTR [ebx+114]
		je c36
		add eax, 34
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -35
	c36:
		test edi, edi
		movsx esi, WORD PTR [ebx+100]
		je c37
		add eax, 35
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -36
	c37:
		test esi, esi
		movsx edi, DWORD PTR [ebx+86]
		je c38
		add eax, 36
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -37
	c38:
		test edi, edi
		movsx esi, WORD PTR [ebx+72]
		je c39
		add eax, 37
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -38
	c39:
		test esi, esi
		movsx edi, DWORD PTR [ebx+58]
		je c40
		add eax, 38
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -39
	c40:
		test edi, edi
		movsx esi, WORD PTR [ebx+44]
		je c41
		add eax, 39
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -40
	c41:
		test esi, esi
		movsx edi, DWORD PTR [ebx+30]
		je c42
		add eax, 40
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -41
	c42:
		test edi, edi
		movsx esi, WORD PTR [ebx+46]
		je c43
		add eax, 41
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -42
	c43:
		test esi, esi
		movsx edi, DWORD PTR [ebx+60]
		je c44
		add eax, 42
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -43
	c44:
		test edi, edi
		movsx esi, WORD PTR [ebx+74]
		je c45
		add eax, 43
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -44
	c45:
		test esi, esi
		movsx edi, DWORD PTR [ebx+88]
		je c46
		add eax, 44
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -45
	c46:
		test edi, edi
		movsx esi, WORD PTR [ebx+102]
		je c47
		add eax, 45
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -46
	c47:


		mov edi, DWORD PTR [ebx+76]		;76 & 78
		mov edx, esi
		mov esi, DWORD PTR [ebx+92]		; 92 & 94
		or edx, edi
		mov edi, DWORD PTR [ebx+104]	;104 & 106
		mov edx, esi
		mov esi, DWORD PTR [ebx+108]	;108 & 110
		or edx, edi
		mov edi, DWORD PTR [ebx+116]	;116 & 118
		mov edx, esi
		mov esi, DWORD PTR [ebx+120]	;120 & 122
		or edx, edi
		mov edi, DWORD PTR [ebx+124]	;124 & 126
		mov edx, esi
		movzx esi, WORD PTR [ebx+62]
		or edx, edi
		movzx edi, WORD PTR [ebx+90]
		mov edx, esi
		or edx, edi
		je c64
		movsx esi, WORD PTR [ebx+102]
			
			
		test esi, esi
		movsx edi, DWORD PTR [ebx+116]
		je c48
		add eax, 46
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -47
	c48:
		test edi, edi
		movsx esi, WORD PTR [ebx+118]
		je c49
		add eax, 47
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -48
	c49:
		test esi, esi
		movsx edi, DWORD PTR [ebx+104]
		je c50
		add eax, 48
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -49
	c50:
		test edi, edi
		movsx esi, WORD PTR [ebx+90]
		je c51
		add eax, 49
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -50
	c51:
		test esi, esi
		movsx edi, DWORD PTR [ebx+76]
		je c52
		add eax, 50
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -51
	c52:
		test edi, edi
		movsx esi, WORD PTR [ebx+62]
		je c53
		add eax, 51
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -52
	c53:
		test esi, esi
		movsx edi, DWORD PTR [ebx+78]
		je c54
		add eax, 52
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -53
	c54:
		test edi, edi
		movsx esi, WORD PTR [ebx+92]
		je c55
		add eax, 53
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -54
	c55:
		test esi, esi
		movsx edi, DWORD PTR [ebx+106]
		je c56
		add eax, 54
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -55
	c56:
		test edi, edi
		movsx esi, WORD PTR [ebx+120]
		je c57
		add eax, 55
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -56
	c57:
		test esi, esi
		movsx edi, DWORD PTR [ebx+122]
		je c58
		add eax, 56
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -57
	c58:
		test edi, edi
		movsx esi, WORD PTR [ebx+108]
		je c59
		add eax, 57
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -58
	c59:
		test esi, esi
		movsx edi, DWORD PTR [ebx+94]
		je c60
		add eax, 58
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -59
	c60:
		test edi, edi
		movsx esi, WORD PTR [ebx+110]
		je c61
		add eax, 59
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -60
	c61:
		test esi, esi
		movsx edi, DWORD PTR [ebx+124]
		je c62
		add eax, 60
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -61
	c62:
		test edi, edi
		movsx esi, WORD PTR [ebx+126]
		je c63
		add eax, 61
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -62
	c63:
		test esi, esi
		je c64
		mov [esp+12],esi
		add eax, 62
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		//jmp block_scanning_end
	c64:
		add ecx, 4
	//block_scanning_end:

		mov eax, ecx

		pop edi;
		pop esi;
		pop edx
		pop ecx
		pop ebx;
		ret 0

	get_vlc_length_eax_esi_eax:
		test esi, esi
		jns is_positive
	//is_negative:
		neg esi
	is_positive:
		test eax, eax
		jne run_is_not_zero
	//run_is_zero:
		movzx eax, WORD PTR [esi*4+(vlc_table2+2)]
		ret
	run_is_not_zero:
		sub eax, 4
		jg run_is_greater_than_4
		sub esi, 17
		jg amp_is_greater_than_17
		imul eax, (17*8)
		mov eax, DWORD PTR [eax+esi*8+(16*8)+(17*8*3)+(vlc_table3+4)]
		ret
	run_is_greater_than_4:
		sub eax, (14-4)
		jg run_is_greater_than_14
		sub esi, 3
		jg amp_is_greater_than_3_run_is_greater_than_4
		imul eax, (3*8)
		mov eax, DWORD PTR [eax+(3*8*9)+esi*8+(2*8)+vlc_table4+4]
		ret
	run_is_greater_than_14:
		movzx eax, WORD PTR [eax*4+(14-1)*4+vlc_table1+2]
		movzx edx, WORD PTR [esi*4+vlc_table2+2]
		add eax, edx
		ret
	amp_is_greater_than_17:
		movzx eax, WORD PTR [eax*4+(4-1)*4+(-1+1)*4+vlc_table1+2]
		movzx edx, WORD PTR [esi*4+(17)*4+vlc_table2+2]
		add eax, edx
		ret
	amp_is_greater_than_3_run_is_greater_than_4:
		movzx eax, WORD PTR [eax*4+9*4+(4-1+1)*4+vlc_table1+2]
		movzx edx, WORD PTR [esi*4+(3)*4+vlc_table2+2]
		add eax, edx
		ret

	get_vlc_length_eax_edi_eax:
		test edi, edi
		jns is_positive_edi
	//is_negative:
		neg edi
	is_positive_edi:
		test eax, eax
		jne run_is_not_zero_edi
	//run_is_zero:
		movzx eax, WORD PTR [edi*4+(vlc_table2+2)]
		ret
	run_is_not_zero_edi:
		sub eax, 4
		jg run_is_greater_than_4_edi
		sub edi, 17
		jg amp_is_greater_than_17_edi
		imul eax, (17*8)
		mov eax, DWORD PTR [eax+edi*8+(16*8)+(17*8*3)+(vlc_table3+4)]
		ret
	run_is_greater_than_4_edi:
		sub eax, (14-4)
		jg run_is_greater_than_14_edi
		sub edi, 3
		jg amp_is_greater_than_3_run_is_greater_than_4_edi
		imul eax, (3*8)
		mov eax, DWORD PTR [eax+(3*8*9)+edi*8+(2*8)+vlc_table4+4]
		ret
	run_is_greater_than_14_edi:
		movzx eax, WORD PTR [eax*4+(14-1)*4+vlc_table1+2]
		movzx edx, WORD PTR [edi*4+vlc_table2+2]
		add eax, edx
		ret
	amp_is_greater_than_17_edi:
		movzx eax, WORD PTR [eax*4+(4-1)*4+(-1+1)*4+vlc_table1+2]
		movzx edx, WORD PTR [edi*4+(17)*4+vlc_table2+2]
		add eax, edx
		ret
	amp_is_greater_than_3_run_is_greater_than_4_edi:
		movzx eax, WORD PTR [eax*4+9*4+(4-1+1)*4+vlc_table1+2]
		movzx edx, WORD PTR [edi*4+(3)*4+vlc_table2+2]
		add eax, edx
		ret

	}
}


unsigned long __declspec(naked) get_vlc_length_248_unrolled_asm( short *block )
{
	__asm {

		push ebx
		push ecx
		push edx
		push esi
		push edi

		mov ebx, DWORD PTR [esp+ 4+(4*5)]	; block
		xor ecx, ecx	; sum of length

		xor eax, eax	; run
		movsx	esi, WORD PTR [ebx+64]

		test esi, esi
		movsx edi, WORD PTR [ebx+ 2]
		je c02
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -1

	c02:
		test edi, edi
		movsx esi, WORD PTR [ebx+66]
		je c03
		add eax, 1
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -2
	c03:
		test esi, esi
		movsx edi, WORD PTR [ebx+16]
		je c04
		add eax, 2
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -3
	c04:
		test edi, edi
		movsx esi, WORD PTR [ebx+80]
		je c05
		add eax, 3
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -4
	c05:
		test esi, esi
		movsx edi, WORD PTR [ebx+ 4]
		je c06
		add eax, 4
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -5
	c06:
		test edi, edi
		movsx esi, WORD PTR [ebx+68]
		je c07
		add eax, 5
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -6
	c07:
		test esi, esi
		movsx edi, WORD PTR [ebx+18]
		je c08
		add eax, 6
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -7
	c08:
		test edi, edi
		movsx esi, WORD PTR [ebx+82]
		je c09
		add eax, 7
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -8
	c09:
		test esi, esi
		movsx edi, WORD PTR [ebx+32]
		je c10
		add eax, 8
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -9
	c10:
		test edi, edi
		movsx esi, WORD PTR [ebx+96]
		je c11
		add eax, 9
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -10
	c11:
		test esi, esi
		movsx edi, WORD PTR [ebx+48]
		je c12
		add eax, 10
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -11
	c12:
		test edi, edi
		movsx esi, WORD PTR [ebx+112]
		je c13
		add eax, 11
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -12
	c13:
		test esi, esi
		movsx edi, WORD PTR [ebx+34]
		je c14
		add eax, 12
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -13
	c14:
		test edi, edi
		movsx esi, WORD PTR [ebx+98]
		je c15
		add eax, 13
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -14
	c15:
		test esi, esi
		movsx edi, WORD PTR [ebx+20]
		je c16
		add eax, 14
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -15
	c16:
		test edi, edi
		movsx esi, WORD PTR [ebx+84]
		je c17
		add eax, 15
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -16
	c17:
		test esi, esi
		movsx edi, WORD PTR [ebx+ 6]
		je c18
		add eax, 16
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -17
	c18:
		test edi, edi
		movsx esi, WORD PTR [ebx+70]
		je c19
		add eax, 17
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -18
	c19:
		test esi, esi
		movsx edi, WORD PTR [ebx+ 8]
		je c20
		add eax, 18
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -19
	c20:
		test edi, edi
		movsx esi, WORD PTR [ebx+72]
		je c21
		add eax, 19
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -20
	c21:
		test esi, esi
		movsx edi, WORD PTR [ebx+22]
		je c22
		add eax, 20
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -21
	c22:
		test edi, edi
		movsx esi, WORD PTR [ebx+86]
		je c23
		add eax, 21
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -22
	c23:
		test esi, esi
		movsx edi, WORD PTR [ebx+36]
		je c24
		add eax, 22
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -23
	c24:
		test edi, edi
		movsx esi, WORD PTR [ebx+100]
		je c25
		add eax, 23
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -24
	c25:
		test esi, esi
		movsx edi, WORD PTR [ebx+50]
		je c26
		add eax, 24
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -25
	c26:
		test edi, edi
		movsx esi, WORD PTR [ebx+114]
		je c27
		add eax, 25
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -26
	c27:
		test esi, esi
		movsx edi, WORD PTR [ebx+52]
		je c28
		add eax, 26
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -27
	c28:
		test edi, edi
		movsx esi, WORD PTR [ebx+116]
		je c29
		add eax, 27
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -28
	c29:
		test esi, esi
		movsx edi, WORD PTR [ebx+38]
		je c30
		add eax, 28
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -29
	c30:
		test edi, edi
		movsx esi, WORD PTR [ebx+102]
		je c31
		add eax, 29
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -30
	c31:
		test esi, esi
		movsx edi, WORD PTR [ebx+24]
		je c32
		add eax, 30
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -31
	c32:
		test edi, edi
		movsx esi, WORD PTR [ebx+88]
		je c33
		add eax, 31
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -32
	c33:
		test esi, esi
		movsx edi, WORD PTR [ebx+10]
		je c34
		add eax, 32
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -33
	c34:
		test edi, edi
		movsx esi, WORD PTR [ebx+74]
		je c35
		add eax, 33
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -34
	c35:
		test esi, esi
		movsx edi, WORD PTR [ebx+12]
		je c36
		add eax, 34
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -35
	c36:
		test edi, edi
		movsx esi, WORD PTR [ebx+76]
		je c37
		add eax, 35
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -36
	c37:
		test esi, esi
		movsx edi, WORD PTR [ebx+26]
		je c38
		add eax, 36
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -37
	c38:
		test edi, edi
		movsx esi, WORD PTR [ebx+90]
		je c39
		add eax, 37
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -38
	c39:
		test esi, esi
		movsx edi, WORD PTR [ebx+40]
		je c40
		add eax, 38
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -39
	c40:
		test edi, edi
		movsx esi, WORD PTR [ebx+104]
		je c41
		add eax, 39
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -40
	c41:
		test esi, esi
		movsx edi, WORD PTR [ebx+54]
		je c42
		add eax, 40
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -41
	c42:
		test edi, edi
		movsx esi, WORD PTR [ebx+118]
		je c43
		add eax, 41
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -42
	c43:
		test esi, esi
		movsx edi, WORD PTR [ebx+56]
		je c44
		add eax, 42
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -43
	c44:
		test edi, edi
		movsx esi, WORD PTR [ebx+120]
		je c45
		add eax, 43
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -44
	c45:
		test esi, esi
		movsx edi, WORD PTR [ebx+42]
		je c46
		add eax, 44
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -45
	c46:
		test edi, edi
		movsx esi, WORD PTR [ebx+106]
		je c47
		add eax, 45
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -46
	c47:
		test esi, esi
		movsx edi, WORD PTR [ebx+28]
		je c48
		add eax, 46
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -47
	c48:
		test edi, edi
		movsx esi, WORD PTR [ebx+92]
		je c49
		add eax, 47
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -48
	c49:
		test esi, esi
		movsx edi, WORD PTR [ebx+14]
		je c50
		add eax, 48
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -49
	c50:
		test edi, edi
		movsx esi, WORD PTR [ebx+78]
		je c51
		add eax, 49
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -50
	c51:
		test esi, esi
		movsx edi, WORD PTR [ebx+30]
		je c52
		add eax, 50
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -51
	c52:
		test edi, edi
		movsx esi, WORD PTR [ebx+94]
		je c53
		add eax, 51
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -52
	c53:
		test esi, esi
		movsx edi, WORD PTR [ebx+44]
		je c54
		add eax, 52
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -53
	c54:
		test edi, edi
		movsx esi, WORD PTR [ebx+108]
		je c55
		add eax, 53
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -54
	c55:
		test esi, esi
		movsx edi, WORD PTR [ebx+58]
		je c56
		add eax, 54
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -55
	c56:
		test edi, edi
		movsx esi, WORD PTR [ebx+122]
		je c57
		add eax, 55
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -56
	c57:
		test esi, esi
		movsx edi, WORD PTR [ebx+60]
		je c58
		add eax, 56
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -57
	c58:
		test edi, edi
		movsx esi, WORD PTR [ebx+124]
		je c59
		add eax, 57
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -58
	c59:
		test esi, esi
		movsx edi, WORD PTR [ebx+46]
		je c60
		add eax, 58
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -59
	c60:
		test edi, edi
		movsx esi, WORD PTR [ebx+110]
		je c61
		add eax, 59
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -60
	c61:
		test esi, esi
		movsx edi, WORD PTR [ebx+62]
		je c62
		add eax, 60
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		mov eax, -61
	c62:
		test edi, edi
		movsx esi, WORD PTR [ebx+126]
		je c63
		add eax, 61
		call get_vlc_length_eax_edi_eax
		add ecx, eax
		mov eax, -62

	c63:
		test esi, esi
		je c64
		mov [esp+12],esi
		add eax, 62
		call get_vlc_length_eax_esi_eax
		add ecx, eax
		//jmp block_scanning_end
	c64:
		add ecx, 4
	//block_scanning_end:

		mov eax, ecx

		pop edi;
		pop esi;
		pop edx
		pop ecx
		pop ebx;
		ret 0

	get_vlc_length_eax_esi_eax:
		test esi, esi
		jns is_positive
	//is_negative:
		neg esi
	is_positive:
		test eax, eax
		jne run_is_not_zero
	//run_is_zero:
		movzx eax, WORD PTR [esi*4+(vlc_table2+2)]
		ret
	run_is_not_zero:
		sub eax, 4
		jg run_is_greater_than_4
		sub esi, 17
		jg amp_is_greater_than_17
		imul eax, (17*8)
		mov eax, DWORD PTR [eax+esi*8+(16*8)+(17*8*3)+(vlc_table3+4)]
		ret
	run_is_greater_than_4:
		sub eax, (14-4)
		jg run_is_greater_than_14
		sub esi, 3
		jg amp_is_greater_than_3_run_is_greater_than_4
		imul eax, (3*8)
		mov eax, DWORD PTR [eax+(3*8*9)+esi*8+(2*8)+vlc_table4+4]
		ret
	run_is_greater_than_14:
		movzx eax, WORD PTR [eax*4+(14-1)*4+vlc_table1+2]
		movzx edx, WORD PTR [esi*4+vlc_table2+2]
		add eax, edx
		ret
	amp_is_greater_than_17:
		movzx eax, WORD PTR [eax*4+(4-1)*4+(-1+1)*4+vlc_table1+2]
		movzx edx, WORD PTR [esi*4+(17)*4+vlc_table2+2]
		add eax, edx
		ret
	amp_is_greater_than_3_run_is_greater_than_4:
		movzx eax, WORD PTR [eax*4+9*4+(4-1+1)*4+vlc_table1+2]
		movzx edx, WORD PTR [esi*4+(3)*4+vlc_table2+2]
		add eax, edx
		ret

	get_vlc_length_eax_edi_eax:
		test edi, edi
		jns is_positive_edi
	//is_negative:
		neg edi
	is_positive_edi:
		test eax, eax
		jne run_is_not_zero_edi
	//run_is_zero:
		movzx eax, WORD PTR [edi*4+(vlc_table2+2)]
		ret
	run_is_not_zero_edi:
		sub eax, 4
		jg run_is_greater_than_4_edi
		sub edi, 17
		jg amp_is_greater_than_17_edi
		imul eax, (17*8)
		mov eax, DWORD PTR [eax+edi*8+(16*8)+(17*8*3)+(vlc_table3+4)]
		ret
	run_is_greater_than_4_edi:
		sub eax, (14-4)
		jg run_is_greater_than_14_edi
		sub edi, 3
		jg amp_is_greater_than_3_run_is_greater_than_4_edi
		imul eax, (3*8)
		mov eax, DWORD PTR [eax+(3*8*9)+edi*8+(2*8)+vlc_table4+4]
		ret
	run_is_greater_than_14_edi:
		movzx eax, WORD PTR [eax*4+(14-1)*4+vlc_table1+2]
		movzx edx, WORD PTR [edi*4+vlc_table2+2]
		add eax, edx
		ret
	amp_is_greater_than_17_edi:
		movzx eax, WORD PTR [eax*4+(4-1)*4+(-1+1)*4+vlc_table1+2]
		movzx edx, WORD PTR [edi*4+(17)*4+vlc_table2+2]
		add eax, edx
		ret
	amp_is_greater_than_3_run_is_greater_than_4_edi:
		movzx eax, WORD PTR [eax*4+9*4+(4-1+1)*4+vlc_table1+2]
		movzx edx, WORD PTR [edi*4+(3)*4+vlc_table2+2]
		add eax, edx
		ret

	}
}




int encode_videosegment( int16_t *QTMB, uint32_t *quantization_number, uint32_t *DCT_mode, uint32_t *class_number, 
						   uint8_t *DIF_block, uint32_t DIF_sequence_number, uint32_t DIF_block_number )
{
	static block_bitbuffer blk_bb[5*6];
	static MB_VLC32buffer MB_VLC32b[5];

	int remain;
	int j;
	int n = 0;
	int16_t *block = QTMB;

	init_bitbuffer_encode( blk_bb, MB_VLC32b, DIF_block );


	for (j=0; j<5; j++)
	{
		int l;
		
		const uint32_t SCT = 4;	// DIF-block-type: video
		const uint32_t Seq = 15;	// Video-sequenznumber: not available
		uint32_t Dseq = DIF_sequence_number;
		uint32_t DBN = DIF_block_number + j;
		const uint32_t STA = 0;	// Macroblock-state
		uint32_t QNO = quantization_number[j];

		((uint32_t*)DIF_block)[j*20] =  (SCT<<29) | 0x10070000 | (Seq<<24) | (Dseq<<20) | (DBN<<8) | (STA<<4) | QNO ;

		for (l=0;l<6;l++)
		{
			uint32_t mo = DCT_mode[n];
			uint32_t c1c0 = class_number[n];
			uint32_t DCI = ( block[0] & 0x1FF );
			uint32_t DCI_m0_c1_c0 = ( (DCI<<3) | (mo<<2) | c1c0 );

			putbits( blk_bb+n, MB_VLC32b+j, DCI_m0_c1_c0, 12 );

			if ( mo==0 )
			{
				block_scanning_88_unrolled_asm( block, blk_bb+n, MB_VLC32b+j );
			}
			else
			{
				block_scanning_248_unrolled_asm( block, blk_bb+n, MB_VLC32b+j );
			}

			block += 64;
			n++;
		}

		// distribute "overflowed" bits on blocks
		if ( movebits( blk_bb+n-6, MB_VLC32b+j ) == 0)
		{
			movebits( blk_bb+n-5, MB_VLC32b+j );
			movebits( blk_bb+n-4, MB_VLC32b+j );
			movebits( blk_bb+n-3, MB_VLC32b+j );
			movebits( blk_bb+n-2, MB_VLC32b+j );
			movebits( blk_bb+n-1, MB_VLC32b+j ); 
		}

	}

	// distribute "overflowed" bits on blocks
	for (j=0; j<5; j++)
	{
		int l;
		n = 0;
		for (l=0; l<4; l++)
		{
			if (j==l) n+=6;
			if( movebits( blk_bb+n+0, MB_VLC32b+j )==1 ) break;
			movebits( blk_bb+n+1, MB_VLC32b+j );
			movebits( blk_bb+n+2, MB_VLC32b+j );
			movebits( blk_bb+n+3, MB_VLC32b+j );
			movebits( blk_bb+n+4, MB_VLC32b+j );
			movebits( blk_bb+n+5, MB_VLC32b+j );
			n+=6;
		}
	}

	byte_swap_block80_asm( DIF_block, 5 );

	remain = 0;
	remain = remain + MB_VLC32b[0].pos - MB_VLC32b[0].data;
	remain = remain + MB_VLC32b[1].pos - MB_VLC32b[1].data;
	remain = remain + MB_VLC32b[2].pos - MB_VLC32b[2].data;
	remain = remain + MB_VLC32b[3].pos - MB_VLC32b[3].data;
	remain = remain + MB_VLC32b[4].pos - MB_VLC32b[4].data;

	return (remain>>1);
}


uint32_t get_videosegment_length( int16_t *QTMB, uint32_t *DCT_mode )
{
	uint32_t vs_length_bits = 5*4*8 + 5*6*12;
	int j;
	int16_t *block = QTMB;

	for (j=0; j<(5*6); j++)
	{
		uint32_t mo = DCT_mode[j];

		if ( mo==0 )
		{
			//uint32_t vs_length_bits_inc = get_vlc_length_88_unrolled( block );
			uint32_t vs_length_bits_inc = get_vlc_length_88_unrolled_asm( block );

			vs_length_bits += vs_length_bits_inc;
		}
		else
		{
			vs_length_bits += get_vlc_length_248_unrolled( block );
		}

		if (vs_length_bits>(5*67*8)) return (vs_length_bits);

		block += 64;
	}

	return (vs_length_bits);
}




